"""Capability Requests management for plugins.

.. versionadded:: 8.0

.. important::

    This is all relatively new. Its usage and documentation is for Sopel core
    development and advanced developers. It is subject to rapid changes
    between versions without much (or any) warning.

    Do **not** build your plugin based on what is here, you do **not** need to.

"""

from __future__ import annotations

import logging
from typing import (
    Optional,
    TYPE_CHECKING,
    Union,
)


if TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from sopel.bot import Sopel, SopelWrapper
    from sopel.plugin import capability, CapabilityNegotiation


LOGGER = logging.getLogger(__name__)


class Manager:
    """Manager of plugins' capability requests.

    Whenever a plugin declares a capability request (through
    :class:`sopel.plugin.capability`), the bot will register the request with
    this manager.

    The bot is responsible to call the manager at the appropriate time, and the
    manager will store requests' state and handle acknowledgement, by following
    this workflow:

    1. the bot will register plugins' requests with the :meth:`register` method
    2. the bot will request the list of available capabilities with the
       ``CAP LS`` subcommand
    3. upon receiving the list, it'll use the :meth:`request_available` method
       to let the manager send the required ``CAP REQ`` messages.
    4. when the server ``ACK`` a request, the bot will call the
       :meth:`acknowledge` method; when the server ``NAK`` a request, the bot
       will call the :meth:`deny` method
    5. once all requests are handled (either directly or after calling the
       :meth:`resume` method), the bot will send the ``CAP END`` message to end
       capability negotiation
    """
    def __init__(self):
        self._registered: dict[
            # CAP REQ :<text>
            tuple[str, ...],
            # mapping (plugin, request + status)
            dict[str, tuple[capability, bool]],
        ] = {}
        self._requested: set[tuple[str, ...]] = set()
        self._acknowledged: set[tuple[str, ...]] = set()
        self._denied: set[tuple[str, ...]] = set()

    # properties

    @property
    def registered(self) -> frozenset[tuple[str, ...]]:
        """Set of registered capability requests.

        Each element is a capability request as a tuple of capability names::

            >>> manager.registered
            {('cap1',), ('cap2', 'cap3')}

        A registered request is a request wanted by a plugin. The request may
        or may not be requested, acknowledged, or denied.
        """
        return frozenset(self._registered.keys())

    @property
    def requested(self) -> frozenset[tuple[str, ...]]:
        """Set of requested capability requests.

        Each element is a capability request as a tuple of capability names::

            >>> manager.requested
            {('cap1',), ('cap2', 'cap3')}

        A requested request is a registered request for which the bot sent
        a ``CAP REQ`` message to the server. The request may or may not be
        acknowledged or denied.

        Only registered requests can be requested.
        """
        return frozenset(self._requested)

    @property
    def acknowledged(self) -> frozenset[tuple[str, ...]]:
        """Set of acknowledged capability requests.

        Each element is a capability request as a tuple of capability names::

            >>> manager.acknowledged
            {('cap1',), ('cap2', 'cap3')}

        An acknowledged request is a registered and requested request for which
        the bot received a ``CAP ACK`` message.

        Only requested requests can be acknowledged.
        """
        return frozenset(self._acknowledged)

    @property
    def denied(self) -> frozenset[tuple[str, ...]]:
        """Set of denied capability requests.

        Each element is a capability request as a tuple of capability names::

            >>> manager.denied
            {('cap1',), ('cap2', 'cap3')}

        A denied request is a registered and requested request for which the
        bot received a ``CAP NAK`` message.

        Only requested requests can be denied.
        """
        return frozenset(self._denied)

    @property
    def is_complete(self) -> bool:
        """Tell if the capability negotiation is complete.

        When capability negotiation is complete, the bot can send ``CAP END``
        to notify the server that negotiation is complete.

        The capability negotiation is complete when all capability requests
        have been either acknowledged or denied successfuly (directly or by
        calling the :meth:`resume` method).
        """
        return all(
            status
            for cap_req, values in self._registered.items()
            for _plugin_capability, status in values.values()
            if cap_req in self._requested
        )

    # tell if registered, requested, acknowledged, or denied

    def is_registered(self, request: Iterable[str]) -> bool:
        """Tell if a capability request is registered.

        :param request: a set of capabilities that form a capability request
                        together; this can be any iterable
        """
        return tuple(sorted(request)) in self._registered

    def is_requested(self, request: Iterable[str]) -> bool:
        """Tell if a capability request is requested.

        :param request: a set of capabilities that form a capability request
                        together; this can be any iterable
        """
        return tuple(sorted(request)) in self._requested

    def is_acknowledged(self, request: Iterable[str]) -> bool:
        """Tell if a capability request is acknowledged.

        :param request: a set of capabilities that form a capability request
                        together; this can be any iterable
        """
        return tuple(sorted(request)) in self._acknowledged

    def is_denied(self, request: Iterable[str]) -> bool:
        """Tell if a capability request is denied.

        :param request: a set of capabilities that form a capability request
                        together; this can be any iterable
        """
        return tuple(sorted(request)) in self._denied

    # register, request, resume, acknowledge, deny, etc.

    def register(self, plugin_name: str, request: capability) -> None:
        """Register a capability ``request`` for ``plugin_name``.

        :param request: the capability request to register for later
        :raise RuntimeError: when the capability request is too long for a
                             single ``CAP REQ`` and ``CAP * ACK``

        Once registered, the capability request can be requested by the bot. A
        registered request appears in :attr:`registered`::

            >>> from sopel import plugin
            >>> request = plugin.capability('cap1')
            >>> manager.register('coretasks', request)
            >>> ('cap1',) in manager.registered
            True
            >>> manager.is_registered(('cap1',))
            True

        It is not, however, directly requested::

            >>> manager.is_requested(('cap1',))
            False

        See :meth:`request_available` to automatically request capabilities
        advertised by the server.

        .. warning::

            Sopel cannot accept a request that is too long, because it does not
            know how to handle a multi-line ACK, and it would not know how to
            call back the appropriate :class:`~sopel.plugin.capability`
            handler.

        """
        cap_req = ' '.join(request.cap_req)
        if len(cap_req.encode('utf-8')) > 500:
            # "CAP * ACK " is 10 bytes, leaving 500 bytes for the capabilities.
            # Sopel cannot allow multi-line requests, as it won't know how to
            # deal properly with multi-line ACK.
            # The spec says a client SHOULD send multiple requests; however
            # the spec also says that a server will ACK or NAK a whole request
            # at once. So technically, multiple REQs are not the same as a
            # single REQ.
            raise RuntimeError('Capability request too long: %s' % cap_req)

        plugin_caps = self._registered.setdefault(request.cap_req, {})
        plugin_caps[plugin_name] = (
            request, False,
        )
        LOGGER.debug('Capability Request registered: %s', request)

    def request_available(
        self,
        bot: Sopel,
        available_capabilities: Iterable[str],
    ) -> None:
        """Request available capabilities.

        :param bot: the bot instance used to send capability requests
        :param available_capabilities: available capabilities

        This sends ``CAP REQ`` commands for requests that can be made, i.e.
        all the requested capabilities (with or without prefix) must be
        available for Sopel to send the request.

        Requests made are stored as requested; others are ignored::

            >>> manager.register('example', 'cap1')
            >>> manager.register('example', 'cap2')
            >>> manager.request_available(bot, ('cap1', 'cap3'))
            >>> manager.is_requested(('cap1',))
            True
            >>> manager.is_requested(('cap2',))
            False
            >>> manager.is_requested(('cap3',))
            False

        .. important::

            The capability request ``('cap1', '-cap2')`` means "enable cap1 and
            disable cap2", and the request will be acknowledged or denied at
            once. If the server doesn't advertise any of these capabilities,
            the client **should not** send a request.

            As a result, Sopel will send a request to enable or disable
            a capability only if it is advertised first. If a request is
            never made, its callback will never be called, because there
            won't be any related ACK/NAK message from the server.

            Plugin authors should not use the request's callback to activate or
            deactivate features, and instead check the bot's capabilities every
            time they need to.

            The only exception to that is when the plugin needs to perform an
            operation while negotiating the capabilities (such as SASL auth).

        """
        # Requesting capabilities when they are available
        capabilities = set(available_capabilities)
        for cap_req, plugin_requests in self._registered.items():
            # request only if all are available
            cap_req_enables = {
                cap.lstrip('-')
                for cap in cap_req
            }
            text = ' '.join(cap_req)
            if cap_req_enables <= capabilities:
                self._registered[cap_req] = {
                    plugin_name: (handler_info[0], False)
                    for plugin_name, handler_info in plugin_requests.items()
                }
                self._requested.add(cap_req)
                LOGGER.debug('Capability negotiation request: "%s".', text)
                bot.write(('CAP', 'REQ'), text=text)
            else:
                LOGGER.debug(
                    'Unable to negotiate capability request: "%s".', text,
                )

    def resume(
        self,
        request: Iterable[str],
        plugin_name: str,
    ) -> tuple[bool, bool]:
        """Resume the registered plugin capability request.

        :return: a 2-value tuple with (was completed, is completed)

        The capability request, for that plugin, will be marked as done,
        and the result will be about the capability negotiation process:

        * was it already completed before the resume?
        * is it completed now?

        If the capability request cannot be found for that plugin, the result
        value remains the same (it stays incomplete or complete)

        .. important::

            When a request's callback returns
            :attr:`~sopel.plugin.CapabilityNegotiation.CONTINUE`, this method
            must be called later (once the plugin has finished its job) or the
            bot will never send the ``CAP END`` command and hang forever.

        .. seealso::

            Plugins can use the method
            :meth:`~sopel.bot.Sopel.resume_capability_negotiation` from the bot
            to resume and automatically send ``CAP END`` when necessary.

        """
        cap_req = tuple(sorted(request))
        was_completed = self.is_complete
        if cap_req not in self._requested:
            return was_completed, was_completed

        handler_info: Optional[tuple[capability, bool]] = self._registered.get(
            cap_req, {},
        ).get(
            plugin_name, None,
        )

        if not handler_info:
            return was_completed, was_completed

        self._registered[cap_req][plugin_name] = (handler_info[0], True)
        return was_completed, self.is_complete

    def acknowledge(
        self,
        bot: SopelWrapper,
        cap_req: tuple[str, ...],
    ) -> Optional[list[tuple[bool, Optional[CapabilityNegotiation]]]]:
        """Acknowledge a capability request and execute handlers.

        :param bot: bot instance to manage the capabilities for
        :param cap_req: the capability request from ``CAP ACK :<cap_req>``
        :return: a list of results and statuses if the capability was requested

        This acknowledges a capability request and executes its callbacks from
        plugins. It returns the result, as a list of 2-value tuples: the first
        value tells if the callback is done with the processing, and the second
        is the returned value.

        If the capability request was denied before, it is now considered
        acknowledged instead.

        If the capability wasn't requested, the result will be ``None``.
        """
        # nothing to acknowledge
        if cap_req not in self._requested:
            LOGGER.debug('Received CAP ACK for an unknown CAP REQ.')
            return None

        # update acknowledged: callbacks may fail, server ACK nonetheless
        self._acknowledged.add(cap_req)

        if cap_req in self._denied:
            self._denied.remove(cap_req)

        # execute callbacks
        return self._callbacks(bot, cap_req, True)

    def deny(
        self,
        bot: SopelWrapper,
        cap_req: tuple[str, ...],
    ) -> Optional[list[tuple[bool, Optional[CapabilityNegotiation]]]]:
        """Deny a capability request and execute handlers.

        :param bot: bot instance to manage the capabilities for
        :param cap_req: the capability request from ``CAP NAK :<cap_req>``

        This denies a capability request and executes its callbacks from
        plugins. It returns the result, as a list of 2-value tuples: the first
        value tells if the callback is done with the processing, and the second
        is the returned value.

        If the capability request was acknowledged before, it is now considered
        denied instead.

        If the capability wasn't requested, the result will be ``None``.
        """
        # nothing to deny
        if cap_req not in self._requested:
            LOGGER.debug('Received CAP NAK for an unknwon CAP REQ.')
            return None

        # update denied: callbacks may fail, server NAK nonetheless
        self._denied.add(cap_req)

        if cap_req in self._acknowledged:
            self._acknowledged.remove(cap_req)

        # execute callbacks
        return self._callbacks(bot, cap_req, False)

    def _callbacks(
        self,
        bot: SopelWrapper,
        cap_req: tuple[str, ...],
        acknowledged: bool,
    ) -> list[tuple[bool, Optional[CapabilityNegotiation]]]:
        # call back request handlers
        plugin_requests: dict[str, tuple[capability, bool]] = self._registered.get(
            cap_req, {},
        )
        return [
            self._callback(plugin_name, handler_info, bot, acknowledged)
            for plugin_name, handler_info in plugin_requests.items()
        ]

    def _callback(
        self,
        plugin_name: str,
        handler_info: tuple[capability, bool],
        bot: SopelWrapper,
        acknowledged: bool,
    ) -> tuple[bool, Optional[CapabilityNegotiation]]:
        handler = handler_info[0]
        is_done, result = handler.callback(bot, acknowledged)
        # update done status in registered
        self._registered[handler.cap_req][plugin_name] = (handler, is_done)
        return (is_done, result)

    def get(
        self,
        cap_req: tuple[str, ...],
        *,
        plugins: Union[list[str], tuple[str, ...], set[str]] = (),
    ) -> Generator[tuple[str, capability], None, None]:
        """Retrieve the registered request handlers for a capability request.

        :param cap_req: the capability request to retrieve handlers for
        :return: yield 2-value tuples with (plugin name, capability)
        """
        plugin_requests = self._registered.get(cap_req, {})
        if plugins:
            yield from (
                (plugin_name, plugin_requests[plugin_name][0])
                for plugin_name in plugins
                if plugin_name in plugin_requests
            )
        else:
            for plugin_name, handler_info in plugin_requests.items():
                yield (plugin_name, handler_info[0])



"""
units.py - Sopel Unit Conversion Plugin
Copyright © 2013, Elad Alfassa, <elad@fedoraproject.org>
Copyright © 2013, Dimitri Molenaars, <tyrope@tyrope.nl>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import re
from typing import Pattern, TYPE_CHECKING

from sopel import plugin


if TYPE_CHECKING:
    from sopel.bot import SopelWrapper
    from sopel.trigger import Trigger


PLUGIN_OUTPUT_PREFIX = '[units] '

find_temp = re.compile(r'(-?[0-9]*\.?[0-9]*)[ °]*(K|C|F)', re.IGNORECASE)
find_length = re.compile(r'([0-9]*\.?[0-9]*)[ ]*(mile[s]?|mi|inch|in|foot|feet|ft|yard[s]?|yd|(?:milli|centi|kilo|)meter[s]?|[mkc]?m|ly|light-year[s]?|au|astronomical unit[s]?|parsec[s]?|pc)', re.IGNORECASE)
find_mass = re.compile(r'([0-9]*\.?[0-9]*)[ ]*(lb|lbm|pound[s]?|ounce|oz|(?:kilo|)gram(?:me|)[s]?|[k]?g)', re.IGNORECASE)


def f_to_c(temp: float) -> float:
    return (float(temp) - 32) * 5 / 9


def c_to_k(temp: float) -> float:
    return temp + 273.15


def c_to_f(temp: float) -> float:
    return (9.0 / 5.0 * temp + 32)


def k_to_c(temp: float) -> float:
    return temp - 273.15


def _extract_source(pattern: Pattern, trigger: Trigger) -> tuple[str, ...]:
    match = pattern.match(trigger.group(2))
    if match:
        return match.groups()
    else:
        raise ValueError("Pattern does not match")


@plugin.command('temp')
@plugin.example('.temp 100F', '37.78°C = 100.00°F = 310.93K')
@plugin.example('.temp 100C', '100.00°C = 212.00°F = 373.15K')
@plugin.example('.temp 100K', '-173.15°C = -279.67°F = 100.00K')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def temperature(bot: SopelWrapper, trigger: Trigger) -> int | None:
    """Convert temperatures"""
    try:
        source = _extract_source(find_temp, trigger)
    except (ValueError, TypeError):
        bot.reply("That's not a valid temperature.")
        return plugin.NOLIMIT
    unit = source[1].upper()
    numeric = float(source[0])
    celsius = 0.0
    if unit == 'C':
        celsius = numeric
    elif unit == 'F':
        celsius = f_to_c(numeric)
    elif unit == 'K':
        celsius = k_to_c(numeric)

    kelvin = c_to_k(celsius)
    fahrenheit = c_to_f(celsius)

    if kelvin <= 0:
        bot.reply("Physically impossible temperature.")
        return None

    bot.say("{:.2f}°C = {:.2f}°F = {:.2f}K".format(
        celsius,
        fahrenheit,
        kelvin,
    ))

    return None


@plugin.command('length', 'distance')
@plugin.example('.distance 3m', '3.00m = 9 feet, 10.11 inches')
@plugin.example('.distance 3km', '3.00km = 1.86 miles')
@plugin.example('.distance 3 miles', '4.83km = 3.00 miles')
@plugin.example('.distance 3 inch', '7.62cm = 3.00 inches')
@plugin.example('.distance 3 feet', '91.44cm = 3 feet, 0.00 inches')
@plugin.example('.distance 3 yards', '2.74m = 9 feet, 0.00 inches')
@plugin.example('.distance 155cm', '1.55m = 5 feet, 1.02 inches')
@plugin.example('.length 3 ly', '28382191417742.40km = 17635876112814.77 miles')
@plugin.example('.length 3 au', '448793612.10km = 278867421.71 miles')
@plugin.example('.length 3 parsec', '92570329129020.20km = 57520535754731.61 miles')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def distance(bot: SopelWrapper, trigger: Trigger) -> int | None:
    """Convert distances"""
    try:
        source = _extract_source(find_length, trigger)
    except (ValueError, TypeError):
        bot.reply("That's not a valid length unit.")
        return plugin.NOLIMIT
    unit = source[1].lower()
    numeric = float(source[0])
    meter = 0.0
    if unit in ("meters", "meter", "m"):
        meter = numeric
    elif unit in ("millimeters", "millimeter", "mm"):
        meter = numeric / 1000
    elif unit in ("kilometers", "kilometer", "km"):
        meter = numeric * 1000
    elif unit in ("miles", "mile", "mi"):
        meter = numeric / 0.00062137
    elif unit in ("inch", "in"):
        meter = numeric / 39.370
    elif unit in ("centimeters", "centimeter", "cm"):
        meter = numeric / 100
    elif unit in ("feet", "foot", "ft"):
        meter = numeric / 3.2808
    elif unit in ("yards", "yard", "yd"):
        meter = numeric / (3.2808 / 3)
    elif unit in ("light-year", "light-years", "ly"):
        meter = numeric * 9460730472580800
    elif unit in ("astronomical unit", "astronomical units", "au"):
        meter = numeric * 149597870700
    elif unit in ("parsec", "parsecs", "pc"):
        meter = numeric * 30856776376340068

    if meter >= 1000:
        metric_part = '{:.2f}km'.format(meter / 1000)
    elif meter < 0.01:
        metric_part = '{:.2f}mm'.format(meter * 1000)
    elif meter < 1:
        metric_part = '{:.2f}cm'.format(meter * 100)
    else:
        metric_part = '{:.2f}m'.format(meter)

    # Shit like this makes me hate being an American.
    inch = meter * 39.37
    foot = int(inch) // 12
    inch = inch - (foot * 12)
    yard = foot // 3
    mile = meter * 0.000621371192

    if yard > 500:
        stupid_part = '{:.2f} miles'.format(mile)
    else:
        parts = []
        if yard >= 100:
            parts.append('{} yards'.format(yard))
            foot -= (yard * 3)

        if foot == 1:
            parts.append('1 foot')
        elif foot != 0:
            parts.append('{:.0f} feet'.format(foot))

        parts.append('{:.2f} inches'.format(inch))

        stupid_part = ', '.join(parts)

    bot.say('{} = {}'.format(metric_part, stupid_part))

    return None


@plugin.command('weight', 'mass')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def mass(bot: SopelWrapper, trigger: Trigger) -> int | None:
    """Convert mass"""
    try:
        source = _extract_source(find_mass, trigger)
    except (ValueError, TypeError):
        bot.reply("That's not a valid mass unit.")
        return plugin.NOLIMIT
    unit = source[1].lower()
    numeric = float(source[0])
    metric = 0.0
    if unit in ("gram", "grams", "gramme", "grammes", "g"):
        metric = numeric
    elif unit in ("kilogram", "kilograms", "kilogramme", "kilogrammes", "kg"):
        metric = numeric * 1000
    elif unit in ("lb", "lbm", "pound", "pounds"):
        metric = numeric * 453.59237
    elif unit in ("oz", "ounce"):
        metric = numeric * 28.35

    if metric >= 1000:
        metric_part = '{:.2f}kg'.format(metric / 1000)
    else:
        metric_part = '{:.2f}g'.format(metric)

    ounce = metric * .035274
    pound = int(ounce) // 16
    ounce = ounce - (pound * 16)

    if pound >= 1:
        stupid_part = '{} {}'.format(pound, 'pound' if pound == 1 else 'pounds')
        if ounce > 0.01:
            stupid_part += ' {:.2f} ounces'.format(ounce)
    else:
        stupid_part = '{:.2f} oz'.format(ounce)

    bot.say('{} = {}'.format(metric_part, stupid_part))

    return None



"""
wikipedia.py - Sopel Wikipedia Plugin
Copyright 2013 Elsie Powell - embolalia.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

from html.parser import HTMLParser
import logging
import re
from urllib.parse import quote, unquote, urlparse

from requests import get

from sopel import plugin
from sopel.config import types


LOGGER = logging.getLogger(__name__)

REDIRECT = re.compile(r'^REDIRECT (.*)')
PLUGIN_OUTPUT_PREFIX = '[wikipedia] '


class WikiParser(HTMLParser):
    NO_CONSUME_TAGS = ('sup', 'style')
    """Tags whose contents should always be ignored.

    These are used in things like inline citations or section "hatnotes", none
    of which are useful output for IRC.
    """

    def __init__(self, section_name):
        HTMLParser.__init__(self)
        self.consume = True
        self.no_consume_depth = 0
        self.is_header = False
        self.section_name = section_name

        self.citations = False
        self.messagebox = False
        self.span_depth = 0
        self.div_depth = 0

        self.result = ''

    def handle_starttag(self, tag, attrs):
        if tag in self.NO_CONSUME_TAGS:
            self.consume = False
            self.no_consume_depth += 1

        elif re.match(r'^h\d$', tag):
            self.is_header = True

        elif tag == 'span':
            if self.span_depth:
                self.span_depth += 1
            else:
                for attr in attrs:  # remove 'edit' tags, and keep track of depth for nested <span> tags
                    if attr[0] == 'class' and 'edit' in attr[1]:
                        self.span_depth += 1

        elif tag == 'div':
            # We want to skip thumbnail text, the table of contents, and section "hatnotes".
            # This also requires tracking div nesting level.
            if self.div_depth:
                self.div_depth += 1
            else:
                for attr in attrs:
                    if attr[0] == 'class' and (
                        'thumb' in attr[1] or 'hatnote' in attr[1] or attr[1] == 'toc'
                    ):
                        self.div_depth += 1
                        break

        elif tag == 'table':
            # Message box templates are what we want to ignore here
            for attr in attrs:
                if (
                    attr[0] == 'class'
                    and any(classname in attr[1].lower() for classname in [
                        # Most of list from https://en.wikipedia.org/wiki/Template:Mbox_templates_see_also
                        'ambox',  # messageboxes on article pages
                        'cmbox',  # messageboxes on category pages
                        'imbox',  # messageboxes on file (image) pages
                        'tmbox',  # messageboxes on talk pages
                        'fmbox',  # header and footer messageboxes
                        'ombox',  # messageboxes on other types of page
                        'mbox',  # for messageboxes that are used in different namespaces and change their presentation accordingly
                        'dmbox',  # for disambiguation messageboxes
                    ])
                ):
                    self.messagebox = True

        elif tag == 'ol':
            for attr in attrs:
                if attr[0] == 'class' and 'references' in attr[1]:
                    self.citations = True   # once we hit citations, we can stop

    def handle_endtag(self, tag):
        if not self.consume and tag in self.NO_CONSUME_TAGS:
            if self.no_consume_depth:
                self.no_consume_depth -= 1
            if not self.no_consume_depth:
                self.consume = True
        if self.is_header and re.match(r'^h\d$', tag):
            self.is_header = False
        if self.span_depth and tag == 'span':
            self.span_depth -= 1
        if self.div_depth and tag == 'div':
            self.div_depth -= 1
        if self.messagebox and tag == 'table':
            self.messagebox = False

    def handle_data(self, data):
        if self.consume and not any([self.citations, self.messagebox, self.span_depth, self.div_depth]):
            if not (self.is_header and data == self.section_name):  # Skip the initial header info only
                self.result += data

    def get_result(self):
        return self.result


class WikipediaSection(types.StaticSection):
    default_lang = types.ValidatedAttribute('default_lang', default='en')
    """The default language to find articles from (same as Wikipedia language subdomain)."""


def setup(bot):
    bot.config.define_section('wikipedia', WikipediaSection)


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | default\\_lang | en | The default language to find articles from (same as Wikipedia language subdomain) |
    """
    config.define_section('wikipedia', WikipediaSection)
    config.wikipedia.configure_setting(
        'default_lang',
        "Enter the default language to find articles from."
    )


def choose_lang(bot, trigger):
    """Determine what language to use for queries based on sender/context."""
    user_lang = bot.db.get_nick_value(trigger.nick, 'wikipedia_lang')
    if user_lang:
        return user_lang

    if not trigger.sender.is_nick():
        channel_lang = bot.db.get_channel_value(trigger.sender, 'wikipedia_lang')
        if channel_lang:
            return channel_lang

    return bot.config.wikipedia.default_lang


def mw_search(server, query, num):
    """Search a MediaWiki site

    Searches the specified MediaWiki server for the given query, and returns
    the specified number of results.
    """
    search_url = ('https://%s/w/api.php?format=json&action=query'
                  '&list=search&srlimit=%d&srprop=timestamp&srwhat=text'
                  '&srsearch=') % (server, num)
    search_url += query
    query = get(search_url).json()
    if 'query' in query:
        query = query['query']['search']
        return [r['title'] for r in query]
    return None


def say_snippet(bot, trigger, server, query, show_url=True, commanded=False):
    page_name = query.replace('_', ' ')
    query = quote(query.replace(' ', '_'))
    url = 'https://{}/wiki/{}'.format(server, query)

    # If the trigger looks like another instance of this plugin, assume it is
    if trigger.startswith(PLUGIN_OUTPUT_PREFIX) and trigger.endswith(' | ' + url):
        return

    try:
        snippet = mw_snippet(server, query)
        # Coalesce repeated whitespace to avoid problems with <math> on MediaWiki
        # see https://github.com/sopel-irc/sopel/issues/2259
        snippet = re.sub(r"\s+", " ", snippet)
    except KeyError:
        msg = 'Error fetching snippet for "{}".'.format(page_name)
        if commanded:
            bot.reply(msg)
        else:
            bot.say(msg)
        return

    msg = '{} | "{}'.format(page_name, snippet)

    trailing = '"'
    if show_url:
        trailing += ' | ' + url

    bot.say(msg, truncation=' […]', trailing=trailing)


def mw_snippet(server, query):
    """Retrieves a snippet of the given page from the given MediaWiki server."""
    snippet_url = ('https://' + server + '/w/api.php?format=json'
                   '&action=query&prop=extracts&exintro&explaintext'
                   '&exchars=500&redirects&titles=')
    snippet_url += query
    snippet = get(snippet_url).json()
    snippet = snippet['query']['pages']

    # For some reason, the API gives the page *number* as the key, so we just
    # grab the first page number in the results.
    snippet = snippet[list(snippet.keys())[0]]

    return snippet['extract']


def say_section(bot, trigger, server, query, section):
    page_name = query.replace('_', ' ')
    query = quote(query.replace(' ', '_'))

    snippet = mw_section(server, query, section)
    if not snippet:
        bot.say('Error fetching section "{}" for page "{}".'.format(section, page_name))
        return

    msg = '{} - {} | "{}"'.format(page_name, section.replace('_', ' '), snippet)
    bot.say(msg, truncation=' […]"')


def mw_section(server, query, section):
    """
    Retrieves a snippet from the specified section from the given page
    on the given server.
    """
    sections_url = ('https://{0}/w/api.php?format=json&redirects'
                    '&action=parse&prop=sections&page={1}'
                    .format(server, query))
    sections = get(sections_url).json()

    fetch_title = section_number = None

    for entry in sections['parse']['sections']:
        if entry['anchor'] == section:
            section_number = entry['index']
            # Needed to handle sections from transcluded pages properly
            # e.g. template documentation (usually pulled in from /doc subpage).
            # One might expect this prop to be nullable because in most cases it
            # will simply repeat the requested page title, but it's always set.
            fetch_title = entry.get('fromtitle')
            break

    if section_number is None or fetch_title is None:
        return None

    snippet_url = ('https://{0}/w/api.php?format=json&redirects'
                   '&action=parse&page={1}&prop=text'
                   '&section={2}').format(server, quote(fetch_title), section_number)

    data = get(snippet_url).json()

    parser = WikiParser(section.replace('_', ' '))
    parser.feed(data['parse']['text']['*'])
    text = parser.get_result()
    text = ' '.join(text.split())  # collapse multiple whitespace chars

    return text


def say_image_description(bot, trigger, server, image):
    desc = mw_image_description(server, image)

    if desc:
        bot.say(desc, truncation=" […]")


def mw_image_description(server, image):
    """Retrieves the description for the given image."""
    params = "&".join([
        "action=query",
        "prop=imageinfo",
        "format=json",
        "indexpageids=1",
        "iiprop=extmetadata",
        "iiextmetadatafilter=ImageDescription",
        "iilimit=1",
        "titles={image}".format(image=image),
    ])
    url = "https://{server}/w/api.php?{params}".format(server=server, params=params)

    response = get(url)
    json = response.json()

    try:
        query_data = json["query"]
        pageids = query_data["pageids"]
        pages = query_data["pages"]

        page = pages[pageids[0]]

        raw_desc = page["imageinfo"][0]["extmetadata"]["ImageDescription"]["value"]

    except LookupError:
        LOGGER.exception("Error getting image description for %r, response was: %r", image, json)
        return None

    # Some descriptions contain markup, use WikiParser to discard that
    parser = WikiParser(image)
    parser.feed(raw_desc)
    desc = parser.get_result()
    desc = ' '.join(desc.split())  # collapse multiple whitespace chars

    return desc


# Matches a Wikipedia link (excluding /File: pages)
@plugin.url(r'https?:\/\/([a-z]+(?:\.m)?\.wikipedia\.org)\/wiki\/((?!File\:)[^ ]+)')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def mw_info(bot, trigger, match=None):
    """Retrieves and outputs a snippet of the linked page."""
    server = match.group(1)
    page_info = urlparse(match.group(0))
    # in Python 3.9+ this can be str.removeprefix() instead, but we're confident that
    # "/wiki/" is at the start of the path anyway since it's part of the pattern
    trim_offset = len("/wiki/")
    article = unquote(page_info.path)[trim_offset:]
    section = unquote(page_info.fragment)

    if article.startswith("Special:"):
        # The MediaWiki query API does not include pages in the Special:
        # namespace, so there's no point bothering when we know this will error
        LOGGER.debug("Ignoring page in Special: namespace")
        return False

    if section:
        if section.startswith('cite_note-'):
            # Don't bother trying to retrieve a section snippet if cite-note is linked
            say_snippet(bot, trigger, server, article, show_url=False)
        elif section.startswith('/media'):
            # gh2316: media fragments are usually images; try to get an image description
            image = section[7:]  # strip '/media' prefix in pre-3.9 friendly way
            say_image_description(bot, trigger, server, image)
        else:
            say_section(bot, trigger, server, article, section)
    else:
        say_snippet(bot, trigger, server, article, show_url=False)


@plugin.command('wikipedia', 'wp')
@plugin.example('.wp San Francisco')
@plugin.output_prefix('[wikipedia] ')
def wikipedia(bot, trigger):
    """Search Wikipedia."""
    if trigger.group(2) is None:
        bot.reply("What do you want me to look up?")
        return plugin.NOLIMIT

    lang = choose_lang(bot, trigger)
    query = trigger.group(2)
    args = re.search(r'^-([a-z]{2,12})\s(.*)', query)
    if args is not None:
        lang = args.group(1)
        query = args.group(2)

    if not query:
        bot.reply('What do you want me to look up?')
        return plugin.NOLIMIT

    if query.startswith("Special:"):
        bot.reply("Sorry, the MediaWiki API doesn't support querying the Special: namespace.")
        return False

    server = lang + '.wikipedia.org'
    query = mw_search(server, query, 1)
    if not query:
        bot.reply("I can't find any results for that.")
        return plugin.NOLIMIT
    else:
        query = query[0]
    say_snippet(bot, trigger, server, query, commanded=True)


@plugin.command('wplang')
@plugin.example('.wplang pl')
def wplang(bot, trigger):
    if not trigger.group(3):
        bot.reply(
            "Your current Wikipedia language is: {}"
            .format(
                bot.db.get_nick_value(
                    trigger.nick, 'wikipedia_lang',
                    bot.config.wikipedia.default_lang)
            )
        )
        return

    bot.db.set_nick_value(trigger.nick, 'wikipedia_lang', trigger.group(3))
    bot.reply(
        "Set your Wikipedia language to: {}"
        .format(trigger.group(3))
    )


@plugin.command('wpclang')
@plugin.example('.wpclang ja')
@plugin.require_chanmsg()
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def wpclang(bot, trigger):
    if not (trigger.admin or bot.channels[trigger.sender.lower()].privileges[trigger.nick.lower()] >= plugin.OP):
        bot.reply("You don't have permission to change this channel's Wikipedia language setting.")
        return plugin.NOLIMIT

    if not trigger.group(3):
        bot.say(
            "{}'s current Wikipedia language is: {}"
            .format(
                trigger.sender,
                bot.db.get_nick_value(
                    trigger.nick, 'wikipedia_lang',
                    bot.config.wikipedia.default_lang)
            )
        )
        return

    bot.db.set_channel_value(trigger.sender, 'wikipedia_lang', trigger.group(3))
    bot.say(
        "Set {}'s Wikipedia language to: {}"
        .format(trigger.sender, trigger.group(3))
    )



"""
wiktionary.py - Sopel Wiktionary Plugin
Copyright 2009, Sean B. Palmer, inamidst.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import re

import requests

from sopel import plugin
from sopel.tools import web

PLUGIN_OUTPUT_PREFIX = '[wiktionary] '

uri = 'https://en.wiktionary.org/w/index.php?title=%s&printable=yes'
r_sup = re.compile(r'<sup[^>]+>.+?</sup>')  # Superscripts that are references only, not ordinal indicators, etc...
r_tag = re.compile(r'<[^>]+>')
r_ul = re.compile(r'(?ims)<ul>.*?</ul>')

# From https://en.wiktionary.org/wiki/Wiktionary:Entry_layout#Part_of_speech
PARTS_OF_SPEECH = [
    # Parts of speech
    'Adjective', 'Adverb', 'Ambiposition', 'Article', 'Circumposition',
    'Classifier', 'Conjunction', 'Contraction', 'Counter', 'Determiner',
    'Ideophone', 'Interjection', 'Noun', 'Numeral', 'Participle', 'Particle',
    'Postposition', 'Preposition', 'Pronoun', 'Proper noun', 'Verb',
    # Morphemes
    'Circumfix', 'Combining form', 'Infix', 'Interfix', 'Prefix', 'Root', 'Suffix',
    # Symbols and characters
    'Diacritical mark', 'Letter', 'Ligature', 'Number', 'Punctuation mark', 'Syllable', 'Symbol',
    # Phrases
    'Phrase', 'Proverb', 'Prepositional phrase',
    # Han characters and language-specific varieties
    'Han character', 'Hanzi', 'Kanji', 'Hanja',
    # Other
    'Romanization',
]


def text(html):
    text = r_sup.sub('', html)  # Remove superscripts that are references from definition
    text = r_tag.sub('', text).strip()
    text = text.replace('\n', ' ')
    text = text.replace('\r', '')
    text = text.replace('(intransitive', '(intr.')
    text = text.replace('(transitive', '(trans.')
    text = web.decode(text)
    return text.strip()


def wikt(word):
    bytes = requests.get(uri % web.quote(word)).text
    bytes = r_ul.sub('', bytes)

    mode = None
    etymology = None
    definitions: dict[str, list[str]] = {}
    for line in bytes.splitlines():
        is_new_mode = False
        if 'id="Etymology' in line:
            mode = 'etymology'
            is_new_mode = True
        else:
            for pos in PARTS_OF_SPEECH:
                if 'id="{}"'.format(pos.replace(' ', '_')) in line:
                    mode = pos.lower()
                    is_new_mode = True
                    break

        if not is_new_mode:
            if (mode == 'etymology') and ('<p>' in line):
                if etymology is not None:
                    # multi-line etymologies do exist (e.g. see "mayhem")
                    etymology += ' ' + text(line)
                else:
                    etymology = text(line)
            # 'id="' can occur in definition lines <li> when <sup> tag is used for references;
            # make sure those are not excluded (e.g. see "abecedarian").
            elif ('id="' in line) and ('<li>' not in line):
                mode = None
            elif (mode is not None) and ('<li>' in line):
                definitions.setdefault(mode, []).append(text(line))

        if '<hr' in line:
            break
    return etymology, definitions


parts = [pos.lower() for pos in PARTS_OF_SPEECH]


def format(result, definitions, number=2):
    for part in parts:
        if part in definitions:
            defs = definitions[part][:number]
            result += u' — {}: '.format(part)
            n = ['%s. %s' % (i + 1, e.strip(' .')) for i, e in enumerate(defs)]
            result += ', '.join(n)
    return result.strip(' .,')


@plugin.command('wt', 'define', 'dict')
@plugin.example('.wt bailiwick', "bailiwick — noun: 1. The district within which a bailie or bailiff has jurisdiction, 2. A person's concern or sphere of operations, their area of skill or authority")
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def wiktionary(bot, trigger):
    """Look up a word on Wiktionary."""
    word = trigger.group(2)
    if word is None:
        bot.reply('You must tell me what to look up!')
        return

    _etymology, definitions = wikt(word)
    if not definitions:
        # Cast word to lower to check in case of mismatched user input
        _etymology, definitions = wikt(word.lower())
        if not definitions:
            bot.reply("Couldn't get any definitions for %s." % word)
            return

    result = format(word, definitions)
    if len(result) < 300:
        result = format(word, definitions, 3)
    if len(result) < 300:
        result = format(word, definitions, 5)

    bot.say(result, truncation=' […]')


@plugin.command('ety')
@plugin.example('.ety bailiwick', "bailiwick: From bailie (“bailiff”) and wick (“dwelling”), from Old English wīc.")
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def wiktionary_ety(bot, trigger):
    """Look up a word's etymology on Wiktionary."""
    word = trigger.group(2)
    if word is None:
        bot.reply('You must give me a word!')
        return

    etymology, _definitions = wikt(word)
    if not etymology:
        bot.reply("Couldn't get the etymology for %s." % word)
        return

    result = "{}: {}".format(word, etymology)

    bot.say(result, truncation=' […]')



"""
countdown.py - Sopel Countdown Plugin
Copyright 2011, Michael Yanovich, yanovich.net
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import annotations

import datetime

from sopel import plugin


@plugin.command('countdown')
@plugin.example('.countdown 2078 09 14')
@plugin.output_prefix('[countdown] ')
def generic_countdown(bot, trigger):
    """Displays a countdown to a given date."""
    text = trigger.group(2)
    if not text:
        bot.reply("Please use correct format: {}countdown 2012 12 21"
                  .format(bot.config.core.help_prefix))
        return plugin.NOLIMIT

    text = trigger.group(2).split()
    if text and (len(text) == 3 and text[0].isdigit() and
                 text[1].isdigit() and text[2].isdigit()):
        try:
            diff = (datetime.datetime(int(text[0]), int(text[1]),
                    int(text[2])) - datetime.datetime.today())
        except Exception:  # TODO: Be specific
            bot.reply("Please use correct format: {}countdown 2012 12 21"
                      .format(bot.config.core.help_prefix))
            return plugin.NOLIMIT
        bot.say(str(diff.days) + " days, " + str(diff.seconds // 3600) +
                " hours and " + str(diff.seconds % 3600 // 60) +
                " minutes until " + text[0] + " " + text[1] + " " + text[2])
    else:
        bot.reply("Please use correct format: {}countdown 2012 12 21"
                  .format(bot.config.core.help_prefix))
        return plugin.NOLIMIT



""":mod:`sopel.irc.backends` defines Sopel's IRC connection handlers.

.. warning::

    This is all internal code, not intended for direct use by plugins. It is
    subject to change between versions, even patch releases, without any
    advance warning.

    Please use the public APIs on :class:`bot <sopel.bot.Sopel>`.

"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

import asyncio
import logging
import signal
import socket
import ssl
import threading
from typing import Any, Optional, TYPE_CHECKING

from .abstract_backends import AbstractIRCBackend


if TYPE_CHECKING:
    from sopel.irc import AbstractBot
    from sopel.trigger import PreTrigger


LOGGER = logging.getLogger(__name__)
QUIT_SIGNALS = [
    getattr(signal, name)
    for name in ['SIGUSR1', 'SIGTERM', 'SIGINT']
    if hasattr(signal, name)
]
RESTART_SIGNALS = [
    getattr(signal, name)
    for name in ['SIGUSR2', 'SIGILL']
    if hasattr(signal, name)
]


class UninitializedBackend(AbstractIRCBackend):
    """IRC Backend shim to use before the bot has started connecting.

    :param bot: an instance of a bot that uses the backend

    This exists to intercept attempts to do "illegal" things before
    connection, like sending messages to IRC before we even have a socket.
    """
    def __init__(
        self,
        bot: AbstractBot,
    ):
        super().__init__(bot)

    def is_connected(self) -> bool:
        """Check if the backend is connected to an IRC server.

        **Always returns False:** This backend type is never connected.

        Jobs (:func:`~.plugin.interval`) or other time-based triggers can be
        invoked before the bot is finished initizalizing, even before it has
        begun the IRC connection process. We need to provide a reliable way for
        those triggers to abort early if they need a connection.

        .. note::

            Your plugin code doesn't need to call this directly; unless your
            plugin does something during connection setup, it should check the
            :attr:`bot.connection_registered <.AbstractBot.connection_registered>`
            flag, which also takes into account whether the IRC connection is
            ready to accept normal commands like PRIVMSG.

        """
        return False

    def on_irc_error(self, pretrigger: PreTrigger) -> None:
        """Dummy IRC error handler.

        Since it should be impossible to receive an error from IRC when there
        is no server connection, this implementation raises an error if it is
        ever called.
        """
        raise RuntimeError("Received error from unconnected backend.")

    def irc_send(self, data: bytes) -> None:
        """Dummy method to send IRC data.

        Since it is impossible to send data to IRC without an IRC connection,
        this implementation raises an error if it is ever called.

        .. note::

            Plugins will likely never need to call this method directly, but
            many public API methods of ``bot`` call it.

            Plugin code that can be triggered by anything that isn't an IRC
            event/message should check the
            :attr:`bot.connection_registered <.AbstractBot.connection_registered>`
            flag before calling any bot method that modifies IRC state.

        """
        raise RuntimeError("Attempt to send data to unconnected backend.")

    def run_forever(self) -> None:
        """Dummy connection setup method.

        Since this implementation is a placeholder and does not actually
        connect to IRC, it raises an error if it is ever called. The bot
        should switch to an appropriate backend type before trying to
        connect; not doing so is a bug.
        """
        raise RuntimeError("Attempt to run dummy backend that cannot connect.")


class AsyncioBackend(AbstractIRCBackend):
    """IRC Backend implementation using :mod:`asyncio`.

    :param bot: an instance of a bot that uses the backend
    :param host: hostname/IP to connect to
    :param port: port to connect to
    :param source_address: optional source address as a tuple of
                           ``(host, port)``
    :param server_timeout: optional time (in seconds) before the backend reach
                           a timeout (defaults to 120s)
    :param ping_interval: optional ping interval (in seconds) between
                          last message received and sending a PING to the
                          server (defaults to ``server_timeout * 0.45``)
    :param use_ssl: if the connection must use SSL/TLS or not
    :param certfile: optional location of the certificates; used when
                     ``use_ssl`` is ``True``
    :param keyfile: optional location to the key file for certificates; used
                    when ``use_ssl`` is ``True`` and ``certfile`` is not
                    ``None``
    :param verify_ssl: if the certificates must be verified; ignored if
                       ``use_ssl`` is not ``True``
    :param ca_certs: optional location to the CA certificates; ignored if
                    ``verify_ssl`` is ``False``
    :param ssl_ciphers: the OpenSSL cipher suites to use
    :param ssl_minimum_version: the lowest SSL/TLS version to accept
    """
    def __init__(
        self,
        bot: AbstractBot,
        host: str,
        port: int,
        source_address: Optional[tuple[str, int]],
        server_timeout: Optional[int] = None,
        ping_interval: Optional[int] = None,
        use_ssl: bool = False,
        certfile: Optional[str] = None,
        keyfile: Optional[str] = None,
        verify_ssl: bool = True,
        ca_certs: Optional[str] = None,
        ssl_ciphers: Optional[list[str]] = None,
        ssl_minimum_version: ssl.TLSVersion = ssl.TLSVersion.TLSv1_2,
        **kwargs: Any,
    ):
        super().__init__(bot)
        # connection parameters
        self._host: str = host
        self._port: int = port
        self._source_address: Optional[tuple[str, int]] = source_address
        self._use_ssl: bool = use_ssl
        self._certfile: Optional[str] = certfile
        self._keyfile: Optional[str] = keyfile
        self._verify_ssl: bool = verify_ssl
        self._ca_certs: Optional[str] = ca_certs
        self._ssl_ciphers: str = ":".join(ssl_ciphers or [])
        self._ssl_minimum_version: ssl.TLSVersion = ssl_minimum_version

        # timeout configuration
        self._server_timeout: float = float(server_timeout or 120)
        self._ping_interval: float = float(
            ping_interval or (self._server_timeout * 0.45)
        )

        # connection flags
        self._connected: bool = False
        self._loop: Optional[asyncio.AbstractEventLoop] = None

        # connection writer & reader
        self._writer: Optional[asyncio.StreamWriter] = None
        self._reader: Optional[asyncio.StreamReader] = None

        # connection tasks
        self._read_task: Optional[asyncio.Task] = None
        self._ping_task: Optional[asyncio.TimerHandle] = None
        self._timeout_task: Optional[asyncio.TimerHandle] = None

    # signal handlers

    def _signal_quit(self) -> None:
        LOGGER.info('Receiving QUIT signal.')
        self.bot.quit('Quit')

    def _signal_restart(self) -> None:
        LOGGER.info('Receiving RESTART signal.')
        self.bot.restart('Restarting')

    # timeout management

    def _ping_callback(self) -> None:
        # simply send a PING
        LOGGER.debug(
            'Sending PING after %0.1fs of inactivity.', self._ping_interval)
        self.send_ping(self._host)

    def _timeout_callback(self) -> None:
        # cancel other tasks
        for task in [self._ping_task, self._read_task]:
            if task is not None:
                task.cancel()
        self._ping_task = None
        self._read_task = None
        # log a warning
        LOGGER.warning(
            'Reached timeout (%0.1fs); closing connection.',
            self._server_timeout,
        )

    def _cancel_timeout_tasks(self) -> None:
        # cancel every timeout tasks (PING & Server Timeout)
        for task in [self._ping_task, self._timeout_task]:
            if task is not None:
                task.cancel()
        self._ping_task = None
        self._timeout_task = None

    def _reset_timeout_tasks(self) -> None:
        # cancel first
        self._cancel_timeout_tasks()
        # then schedule again
        loop = asyncio.get_running_loop()
        self._ping_task = loop.call_later(
            self._ping_interval, self._ping_callback,
        )
        self._timeout_task = loop.call_later(
            self._server_timeout, self._timeout_callback,
        )

    # backend interface

    def is_connected(self) -> bool:
        return self._connected

    def on_irc_error(self, pretrigger: PreTrigger) -> None:
        LOGGER.warning('Error received from server: %s', pretrigger.text)

    def irc_send(self, data: bytes) -> None:
        if self._loop is None:
            raise RuntimeError('EventLoop not initialized.')

        if threading.current_thread() is threading.main_thread():
            self._loop.create_task(self.send(data))
        else:
            asyncio.run_coroutine_threadsafe(self.send(data), self._loop)

    # read/write

    async def send(self, data: bytes) -> None:
        """Send ``data`` through the writer."""
        if self._writer is None:
            raise RuntimeError(
                'Writer not initialized. '
                'Are you sure the backend is running?')

        try:
            self._writer.write(data)
            await self._writer.drain()
        except asyncio.CancelledError:
            LOGGER.debug('Writer was cancelled')

    async def read_forever(self) -> None:
        """Main reading loop of the backend.

        This listens to the reader for an incoming IRC line, decodes the data,
        and passes it to
        :meth:`bot.on_message(data) <sopel.irc.AbstractBot.on_message>`, until
        the reader reaches the EOF (i.e. connection closed).

        It manages connection timeouts by scheduling two tasks:

        * a PING task, that will send a PING to the server as defined by
          the ping interval (from the configuration)
        * a Timeout task, that will stop the bot if it reaches the timeout

        Whenever a message is received, both tasks are cancelled and
        rescheduled.

        When the connection is closed, the reader will reach EOF, and return
        an empty string, which in turn will end the coroutine.

        .. seealso::

            The :meth:`~.decode_line` method is used to decode the IRC line
            from :class:`bytes` to :class:`str`.

        """
        if self._reader is None:
            raise RuntimeError(
                'Reader not initialized. '
                'Are you sure the backend is running?')

        # cancel timeout tasks
        self._cancel_timeout_tasks()

        # loop forever until EOF
        while not self._reader.at_eof():
            try:
                line: bytes = await self._reader.readuntil(separator=b'\r\n')
            except asyncio.IncompleteReadError as err:
                LOGGER.warning('Receiving partial message from IRC.')
                line = err.partial
            except asyncio.LimitOverrunError:
                LOGGER.exception('Unable to read from IRC server.')
                break

            # connection is active: reset timeout tasks
            self._reset_timeout_tasks()

            # check content
            if not line:
                LOGGER.debug('No data received.')
                continue

            # decode content to unicode
            try:
                data: str = self.decode_line(line)
            except ValueError:
                LOGGER.error('Unable to decode line from IRC server: %r', line)
                continue

            # use bot's callbacks
            try:
                self.bot.log_raw(data, '<<')
                self.bot.on_message(data)
            except Exception:
                LOGGER.exception('Unexpected exception on message handling.')
                LOGGER.warning('Stopping the backend after error.')
                break

        # cancel timeout tasks when reading loop ends
        self._cancel_timeout_tasks()

    # run & connection

    def get_connection_kwargs(self) -> dict:
        """Return the keyword arguments required to initiate connection."""
        ssl_context: Optional[ssl.SSLContext] = None

        if self._use_ssl:
            ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            ssl_context.minimum_version = self._ssl_minimum_version
            ssl_context.set_ciphers(self._ssl_ciphers)
            if self._certfile is not None:
                # load_cert_chain requires a certfile (cannot be None)
                ssl_context.load_cert_chain(
                    certfile=self._certfile,
                    keyfile=self._keyfile,
                )

            if self._verify_ssl and self._ca_certs is not None:
                ssl_context.load_verify_locations(self._ca_certs)
            elif not self._verify_ssl:
                # deactivate SSL verification for hostname & certificate
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE

        return {
            'host': self._host,
            'port': self._port,
            'ssl': ssl_context,
            'local_addr': self._source_address,
        }

    async def _connect_to_server(
        self, **connection_kwargs: Any,
    ) -> tuple[
        Optional[asyncio.StreamReader],
        Optional[asyncio.StreamWriter],
    ]:
        reader: Optional[asyncio.StreamReader] = None
        writer: Optional[asyncio.StreamWriter] = None

        # open connection
        try:
            reader, writer = await asyncio.open_connection(
                **connection_kwargs,
            )

        # SSL Errors (certificate verification and generic SSL errors)
        except ssl.SSLCertVerificationError as err:
            LOGGER.error(
                'Unable to connect due to '
                'SSL certificate verification failure: %s',
                err,
            )
            self.log_exception()
            # tell the bot to quit without restart
            self.bot.hasquit = True
            self.bot.wantsrestart = False
        except ssl.SSLError as err:
            LOGGER.error('Unable to connect due to an SSL error: %s', err)
            self.log_exception()
            # tell the bot to quit without restart
            self.bot.hasquit = True
            self.bot.wantsrestart = False

        # Specific connection error (invalid address and timeout)
        except socket.gaierror as err:
            LOGGER.error(
                'Unable to connect due to invalid IRC server address: %s',
                err,
            )
            LOGGER.error(
                'You should verify that "%s:%s" is the correct address '
                'to connect to the IRC server.',
                connection_kwargs.get('host'),
                connection_kwargs.get('port'),
            )
            self.log_exception()
            # tell the bot to quit without restart
            self.bot.hasquit = True
            self.bot.wantsrestart = False
        except TimeoutError as err:
            LOGGER.error('Unable to connect due to a timeout: %s', err)
            self.log_exception()
            # tell the bot to quit without restart
            self.bot.hasquit = True
            self.bot.wantsrestart = False

        # Generic connection error
        except ConnectionError as err:
            LOGGER.error('Unable to connect: %s', err)
            self.log_exception()
            # tell the bot to quit without restart
            self.bot.hasquit = True
            self.bot.wantsrestart = False

        # Generic OSError (used for any unspecific connection error)
        except OSError as err:
            LOGGER.error('Unable to connect: %s', err)
            LOGGER.error(
                'You should verify that "%s:%s" is the correct address '
                'to connect to the IRC server.',
                connection_kwargs.get('host'),
                connection_kwargs.get('port'),
            )
            self.log_exception()
            # tell the bot to quit without restart
            self.bot.hasquit = True
            self.bot.wantsrestart = False

        # Unexpected error
        except Exception as err:
            LOGGER.error(
                'Unable to connect due to an unexpected error: %s',
                err,
            )
            self.log_exception()
            # until there is a way to prevent an infinite loop of connection
            # error and reconnect, we have to tell the bot to quit here
            # TODO: prevent infinite connection failure loop
            self.bot.hasquit = True
            self.bot.wantsrestart = False

        return reader, writer

    async def _run_forever(self) -> None:
        self._loop = asyncio.get_running_loop()
        connection_kwargs = self.get_connection_kwargs()

        # register signal handlers
        for quit_signal in QUIT_SIGNALS:
            self._loop.add_signal_handler(quit_signal, self._signal_quit)
        for restart_signal in RESTART_SIGNALS:
            self._loop.add_signal_handler(restart_signal, self._signal_restart)

        # connect to socket
        LOGGER.debug('Attempt connection.')
        self._reader, self._writer = await self._connect_to_server(
            **connection_kwargs
        )
        if not self._reader or not self._writer:
            LOGGER.debug('Connection attempt failed.')
            return

        # on socket connection
        LOGGER.debug('Connection registered.')
        self._connected = True
        self.bot.on_connect()

        # read forever
        LOGGER.debug('Waiting for messages...')
        self._read_task = asyncio.create_task(self.read_forever())
        try:
            await self._read_task

        # task was cancelled, i.e. another exception is responsible for that
        except asyncio.CancelledError:
            LOGGER.debug('Read task was cancelled.')

        # connection reset requires a log, but no exception log
        except ConnectionResetError as err:
            LOGGER.error('Connection reset on read: %s', err)

        # generic (connection) error requires a specific exception log
        except ConnectionError as err:
            LOGGER.error('Connection error on read: %s', err)
            self.log_exception()

        except Exception as err:
            LOGGER.error('Unexpected error on read: %s', err)
            self.log_exception()

        # task done (connection closed without error)
        else:
            LOGGER.debug('Reader received EOF.')

        # on socket disconnection
        self._connected = False

        # cancel timeout tasks
        self._cancel_timeout_tasks()

        # nothing to read anymore
        LOGGER.debug('Shutting down writer.')
        self._writer.close()
        try:
            await self._writer.wait_closed()

        # task was cancelled, i.e. another exception is responsible for that
        except asyncio.CancelledError:
            LOGGER.debug('Writer task was cancelled.')

        # connection reset happened before on the read task
        except ConnectionResetError as err:
            LOGGER.debug('Connection reset while closing: %s', err)

        # generic (connection) error requires a specific exception log
        except ConnectionError as err:
            LOGGER.error('Connection error while closing: %s', err)
            self.log_exception()

        except Exception:
            LOGGER.error('Unexpected error while shutting down socket.')
            self.log_exception()

        LOGGER.debug('All clear, exiting now.')

    def run_forever(self) -> None:
        """Run forever."""
        LOGGER.debug('Running forever.')
        asyncio.run(self._run_forever())
        LOGGER.info('Connection backend stopped.')
        self.bot.on_close()



"""
Sopel is a simple, easy-to-use, open-source IRC utility bot, written in Python.

It’s designed to be easy to use, easy to run, and easy to extend.
"""
#
# Copyright 2008, Sean B. Palmer, inamidst.com
# Copyright 2012, Elsie Powell, http://embolalia.com
# Copyright 2012, Elad Alfassa <elad@fedoraproject.org>
#
# Licensed under the Eiffel Forum License 2.

from __future__ import annotations

from collections import namedtuple
import importlib.metadata
import locale
import re
import sys


__all__ = [
    'bot',
    'config',
    'db',
    'formatting',
    'irc',
    'loader',
    'logger',
    'module',  # deprecated in 7.1, removed in 9.0
    'plugin',
    'tools',
    'trigger',
    'version_info',
]

loc = locale.getlocale()
if not loc[1] or ('UTF-8' not in loc[1] and 'utf8' not in loc[1]):
    print('Warning: Running with a non-UTF8 locale. If you see strange '
          'encoding errors, try setting the LC_ALL environment variable to '
          'something like "en_US.UTF-8".', file=sys.stderr)


__version__ = importlib.metadata.version('sopel')


def _version_info(version=__version__):
    regex = re.compile(r'(\d+)\.(\d+)\.(\d+)(?:[\-\.]?(a|b|rc)(\d+))?.*')
    version_match = regex.match(version)

    if version_match is None:
        raise RuntimeError("Can't parse version number!")

    version_groups = version_match.groups()
    major, minor, micro = (int(piece) for piece in version_groups[0:3])
    level = version_groups[3]
    serial = int(version_groups[4] or 0)
    if level == 'a':
        level = 'alpha'
    elif level == 'b':
        level = 'beta'
    elif level == 'rc':
        level = 'candidate'
    elif not level and version_groups[4] is None:
        level = 'final'
    else:
        level = 'alpha'

    VersionInfo = namedtuple('VersionInfo',
                             'major, minor, micro, releaselevel, serial')
    return VersionInfo(major, minor, micro, level, serial)


version_info = _version_info()



"""Tools to help safely do calculations from user input

.. versionadded:: 5.3
.. note::

    Most of this is internal machinery. :func:`eval_equation` is the "public"
    part, used by Sopel's built-in ``calc`` plugin.

"""
from __future__ import annotations

import ast
import operator
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Optional

__all__ = ['eval_equation']


class ExpressionEvaluator:
    """A generic class for evaluating limited forms of Python expressions.

    :param dict bin_ops: optional; see notes
    :param dict unary_ops: optional; see notes

    Instances can pass ``binary_ops`` and ``unary_ops`` arguments with dicts of
    the form ``{ast.Node, function}``. When the :class:`ast.Node <ast.AST>` used
    as key is found, it will be evaluated using the given ``function``.
    """

    class Error(Exception):
        """Internal exception type for :class:`ExpressionEvaluator`\\s."""

    def __init__(
        self,
        bin_ops: Optional[dict[type[ast.operator], Callable]] = None,
        unary_ops: Optional[dict[type[ast.unaryop], Callable]] = None
    ):
        self.binary_ops = bin_ops or {}
        self.unary_ops = unary_ops or {}

    def __call__(
        self,
        expression_str: str,
        timeout: float = 5.0,
    ) -> int | float:
        """Evaluate a Python expression and return the result.

        :param expression_str: the expression to evaluate
        :param timeout: timeout for processing the expression, in seconds
        :raise SyntaxError: if the given ``expression_str`` is not a valid
                            Python statement
        :raise ExpressionEvaluator.Error: if the instance of
            :class:`ExpressionEvaluator` does not have a handler for the
            :class:`ast.Node <ast.AST>`
        """
        ast_expression = ast.parse(expression_str, mode='eval')
        return self._eval_node(ast_expression.body, time.time() + timeout)

    def _eval_node(self, node: ast.AST, timeout: float) -> float:
        """Recursively evaluate the given :class:`ast.Node <ast.AST>`.

        :param node: the AST node to evaluate
        :param timeout: how long the expression is allowed to process before
                        timing out and failing, in seconds
        :raise ExpressionEvaluator.Error: if it can't handle the ``node``

        Uses :attr:`self.binary_ops` and :attr:`self.unary_ops` for the
        implementation.

        A subclass could overwrite this to handle more nodes, calling it only
        for nodes it does not implement itself.
        """
        if isinstance(node, ast.Constant):
            if not isinstance(node.value, (int, float)):
                raise ExpressionEvaluator.Error(
                    "'{}' values are not supported".format(
                        type(node.value).__name__,
                    )
                )

            return node.value

        elif isinstance(node, ast.BinOp):
            if type(node.op) not in self.binary_ops:
                raise ExpressionEvaluator.Error(
                    "Unsupported binary operator '{}'".format(
                        type(node.op).__name__,
                    )
                )

            left = self._eval_node(node.left, timeout)
            right = self._eval_node(node.right, timeout)
            if time.time() > timeout:
                raise ExpressionEvaluator.Error(
                    "Time for evaluating expression ran out.")
            return self.binary_ops[type(node.op)](left, right)

        elif isinstance(node, ast.UnaryOp):
            if type(node.op) not in self.unary_ops:
                raise ExpressionEvaluator.Error(
                    "Unsupported unary operator '{}'".format(
                        type(node.op).__name__,
                    )
                )

            operand = self._eval_node(node.operand, timeout)
            if time.time() > timeout:
                raise ExpressionEvaluator.Error(
                    "Time for evaluating expression ran out.")
            return self.unary_ops[type(node.op)](operand)

        raise ExpressionEvaluator.Error(
            "Node type '{}' is not supported.".format(
                type(node).__name__,
            )
        )


def guarded_mul(left: float, right: float) -> float:
    """Multiply two values, guarding against overly large inputs.

    :param left: the left operand
    :param right: the right operand
    :raise ValueError: if the inputs are too large to handle safely
    """
    # Only handle ints because floats will overflow anyway.
    if not isinstance(left, int) or not isinstance(right, int):
        pass
    elif left in (0, 1) or right in (0, 1):
        # Ignore trivial cases.
        pass
    elif left.bit_length() + right.bit_length() > 664386:
        # 664386 is the number of bits (10**100000)**2 has, which is instant on
        # my laptop, while (10**1000000)**2 has a noticeable delay. It could
        # certainly be improved.
        raise ValueError(
            "Value is too large to be handled in limited time and memory.")

    return operator.mul(left, right)


def pow_complexity(num: int, exp: int) -> float:
    """Estimate the worst case time :func:`pow` takes to calculate.

    :param num: base
    :param exp: exponent

    This function is based on experimental data from the time it takes to
    calculate ``num**exp`` in 32-bit CPython 2.7.6 on an Intel Core i7-2670QM
    laptop running Windows.

    It tries to implement this surface: x=exp, y=num

    +----+------+------+------+------+-------+-------+-------+-------+-------+
    |    |  1e5 |  2e5 |  3e5 |  4e5 |  5e5  |  6e5  |  7e5  |  8e5  |  9e5  |
    +====+======+======+======+======+=======+=======+=======+=======+=======+
    | e1 | 0.03 | 0.09 | 0.16 | 0.25 |  0.35 |  0.46 |  0.60 |  0.73 |  0.88 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e2 | 0.08 | 0.24 | 0.46 | 0.73 |  1.03 |  1.40 |  1.80 |  2.21 |  2.63 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e3 | 0.15 | 0.46 | 0.87 | 1.39 |  1.99 |  2.63 |  3.35 |  4.18 |  5.15 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e4 | 0.24 | 0.73 | 1.39 | 2.20 |  3.11 |  4.18 |  5.39 |  6.59 |  7.88 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e5 | 0.34 | 1.03 | 2.00 | 3.12 |  4.48 |  5.97 |  7.56 |  9.37 | 11.34 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e6 | 0.46 | 1.39 | 2.62 | 4.16 |  5.97 |  7.86 | 10.09 | 12.56 | 15.39 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e7 | 0.60 | 1.79 | 3.34 | 5.39 |  7.60 | 10.16 | 13.00 | 16.23 | 19.44 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e8 | 0.73 | 2.20 | 4.18 | 6.60 |  9.37 | 12.60 | 16.26 | 19.83 | 23.70 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e9 | 0.87 | 2.62 | 5.15 | 7.93 | 11.34 | 15.44 | 19.40 | 23.66 | 28.58 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+

    For powers of 2 it tries to implement this surface:

    +---+------+------+------+------+------+------+------+------+------+
    |   |  1e7 |  2e7 |  3e7 |  4e7 |  5e7 |  6e7 |  7e7 |  8e7 |  9e7 |
    +===+======+======+======+======+======+======+======+======+======+
    | 1 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 |
    +---+------+------+------+------+------+------+------+------+------+
    | 2 | 0.21 | 0.44 | 0.71 | 0.92 | 1.20 | 1.49 | 1.66 | 1.95 | 2.23 |
    +---+------+------+------+------+------+------+------+------+------+
    | 4 | 0.43 | 0.91 | 1.49 | 1.96 | 2.50 | 3.13 | 3.54 | 4.10 | 4.77 |
    +---+------+------+------+------+------+------+------+------+------+
    | 8 | 0.70 | 1.50 | 2.24 | 3.16 | 3.83 | 4.66 | 5.58 | 6.56 | 7.67 |
    +---+------+------+------+------+------+------+------+------+------+

    The function numbers were selected by starting with the theoretical
    complexity of ``exp * log2(num)**2`` and fiddling with the exponents
    until it more or less matched with the table.

    Because this function is based on a limited set of data it might not give
    accurate results outside these boundaries. The results derived from large
    ``num`` and ``exp`` were quite accurate for small ``num`` and very large
    ``exp`` though, except when ``num`` was a power of 2.
    """
    if num in (0, 1) or exp in (0, 1):
        return 0
    elif (num & (num - 1)) == 0:
        # For powers of 2 the scaling is a bit different.
        return exp ** 1.092 * num.bit_length() ** 1.65 / 623212911.121
    else:
        return exp ** 1.590 * num.bit_length() ** 1.73 / 36864057619.3


def guarded_pow(num: float, exp: float) -> float:
    """Raise a number to a power, guarding against overly large inputs.

    :param num: base
    :param exp: exponent
    :raise ValueError: if the inputs are too large to handle safely
    """
    # Only handle ints because floats will overflow anyway.
    if not isinstance(num, int) or not isinstance(exp, int):
        pass
    elif pow_complexity(num, exp) < 0.5:
        # Value 0.5 is arbitrary and based on an estimated runtime of 0.5s
        # on a fairly decent laptop processor.
        pass
    else:
        raise ValueError("Pow expression too complex to calculate.")

    return operator.pow(num, exp)


class EquationEvaluator(ExpressionEvaluator):
    """Specific subclass of :class:`ExpressionEvaluator` for simple math

    This presets the allowed operators to safeguard against user input that
    could try to do things that will adversely affect the running bot, while
    still letting users pass arbitrary mathematical expressions using the
    available (mostly arithmetic) operators.
    """
    __bin_ops: dict[type[ast.operator], Callable] = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: guarded_mul,
        ast.Div: operator.truediv,
        ast.Pow: guarded_pow,
        ast.Mod: operator.mod,
        ast.FloorDiv: operator.floordiv,
        ast.BitXor: guarded_pow
    }
    __unary_ops: dict[type[ast.unaryop], Callable] = {
        ast.USub: operator.neg,
        ast.UAdd: operator.pos,
    }

    def __init__(self):
        ExpressionEvaluator.__init__(
            self,
            bin_ops=self.__bin_ops,
            unary_ops=self.__unary_ops
        )

    def __call__(
        self,
        expression_str: str,
        timeout: float = 5.0,
    ) -> float:
        result = ExpressionEvaluator.__call__(self, expression_str, timeout)

        # This wrapper is here so additional sanity checks could be done
        # on the result of the eval, but currently none are done.

        return result


eval_equation = EquationEvaluator()
"""Evaluates a Python equation expression and returns the result.

:param str equation: the equation to evaluate
:param timeout: optional timeout value
:type timeout: int or float

Supports addition (+), subtraction (-), multiplication (*), division (/),
power (**) and modulo (%).
"""



""":mod:`sopel.irc` is the core IRC module for Sopel.

This sub-package contains everything that is related to the IRC protocol
(connection, commands, abstract client, etc.) and that can be used to implement
the Sopel bot.

In particular, it defines the interface for the IRC backend
(:class:`~sopel.irc.abstract_backends.AbstractIRCBackend`), and the
interface for the bot itself (:class:`~sopel.irc.AbstractBot`).

.. warning::

    This is all internal code, not intended for direct use by plugins. It is
    subject to change between versions, even patch releases, without any
    advance warning.

    Please use the public APIs on :class:`bot <sopel.bot.Sopel>`.

.. important::

    When working on core IRC protocol related features, consult protocol
    documentation at https://modern.ircdocs.horse/

"""
# Copyright 2008, Sean B. Palmer, inamidst.com
# Copyright 2012, Elsie Powell, http://embolalia.com
# Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

import abc
from collections import deque
from datetime import datetime, timezone
import logging
import os
import threading
import time
from typing import (
    Any,
    TYPE_CHECKING,
)

from sopel import tools, trigger
from sopel.lifecycle import deprecated
from sopel.tools import identifiers, memories
from .backends import AsyncioBackend, UninitializedBackend
from .capabilities import Capabilities
from .isupport import ISupport

if TYPE_CHECKING:
    from collections.abc import Iterable
    from sopel.config import Config
    from .abstract_backends import AbstractIRCBackend
    from .utils import MyInfo


__all__ = ['abstract_backends', 'backends', 'utils']

LOGGER = logging.getLogger(__name__)
ERR_BACKEND_NOT_INITIALIZED = 'Backend not initialized; is the bot running?'


class AbstractBot(abc.ABC):
    """Abstract definition of Sopel's interface."""
    def __init__(self, settings: Config):
        # private properties: access as read-only properties
        self._user: str = settings.core.user
        self._name: str = settings.core.name
        self._isupport = ISupport()
        self._capabilities = Capabilities()
        self._myinfo: MyInfo | None = None
        self._nick: identifiers.Identifier = self.make_identifier(
            settings.core.nick)

        self.backend: AbstractIRCBackend = UninitializedBackend(self)
        """IRC Connection Backend."""
        self._connection_registered = threading.Event()
        """Flag stating whether the IRC connection is registered yet."""
        self.settings = settings
        """The bot's settings.

        .. versionadded:: 7.0
        """

        # internal machinery
        self.sending = threading.RLock()
        self.last_error_timestamp: datetime | None = None
        self.error_count = 0
        self.stack: dict[identifiers.Identifier, dict[str, Any]] = {}
        self.hasquit = False
        self.wantsrestart = False
        self.last_raw_line = ''  # last raw line received

    @property
    def connection_registered(self) -> bool:
        """Whether the IRC connection is registered.

        This is a property so it can accurately reflect not only the socket
        state (connection to IRC server), but also whether the connection is
        ready to accept "normal" IRC commands.

        Before registration is completed, only a very limited set of commands
        are allowed to be used. Sopel itself takes care of these, so plugins
        will be more concerned with whether they are allowed to use methods
        like :meth:`say` yet.
        """
        return (
            self.backend is not None
            and self.backend.is_connected()
            and self._connection_registered.is_set())

    @property
    def nick(self) -> identifiers.Identifier:
        """Sopel's current nick.

        Changing this while Sopel is running is unsupported and can result in
        undefined behavior.
        """
        return self._nick

    @property
    def user(self) -> str:
        """Sopel's user/ident."""
        return self._user

    @property
    def name(self) -> str:
        """Sopel's "real name", as used for WHOIS responses."""
        return self._name

    @property
    def config(self) -> Config:
        """The :class:`sopel.config.Config` for the current Sopel instance."""
        # TODO: Deprecate config, replaced by settings
        return self.settings

    @property
    def capabilities(self) -> Capabilities:
        """Capabilities negotiated with the server.

        .. versionadded:: 8.0
        """
        return self._capabilities

    @property
    @deprecated(
        reason='Capability handling has been rewritten. '
        'Use `bot.capabilities.is_enabled()` or `bot.capabilities.enabled` instead.',
        version='8.0',
        warning_in='8.1',
        removed_in='9.0',
    )
    def enabled_capabilities(self) -> set[str]:
        """A set containing the IRCv3 capabilities that the bot has enabled.

        .. deprecated:: 8.0

            Enabled server capabilities are now managed by
            :attr:`bot.capabilities <capabilities>` and its various methods and
            attributes:

            * use :meth:`bot.capabilities.is_enabled() <sopel.irc.capabilities.Capabilities.is_enabled>`
              to check if a capability is enabled
            * use :attr:`bot.capabilities.enabled <sopel.irc.capabilities.Capabilities.enabled>`
              for a list of enabled capabilities

            Will be removed in Sopel 9.

        """
        return set(self._capabilities.enabled)

    @property
    @deprecated(
        reason='Capability handling has been rewritten. '
        'Use `bot.capabilities.is_available()` or `bot.capabilities.available` instead.',
        version='8.0',
        warning_in='8.1',
        removed_in='9.0',
    )
    def server_capabilities(self) -> dict[str, str | None]:
        """A dict mapping supported IRCv3 capabilities to their options.

        For example, if the server specifies the capability ``sasl=EXTERNAL``,
        it will be here as ``{"sasl": "EXTERNAL"}``. Capabilities specified
        without any options will have ``None`` as the value.

        For servers that do not support IRCv3, this will be an empty set.

        .. deprecated:: 8.0

            Enabled server capabilities are now managed by
            :attr:`bot.capabilities <capabilities>` and its various methods and
            attributes:

            * use :meth:`bot.capabilities.is_available() <sopel.irc.capabilities.Capabilities.is_available>`
              to check if a capability is available
            * use :attr:`bot.capabilities.available <sopel.irc.capabilities.Capabilities.available>`
              for a list of available capabilities and their parameters

            Will be removed in Sopel 9.

        """
        return self._capabilities.available

    @property
    def isupport(self) -> ISupport:
        """Features advertised by the server.

        .. versionadded:: 7.0
        """
        return self._isupport

    @property
    def myinfo(self) -> MyInfo:
        """Server/network information.

        .. versionadded:: 7.0
        """
        if self._myinfo is None:
            raise AttributeError('myinfo')
        return self._myinfo

    @property
    @abc.abstractmethod
    def hostmask(self) -> str | None:
        """The bot's hostmask."""

    # Utility

    def make_identifier(self, name: str) -> identifiers.Identifier:
        """Instantiate an Identifier using the bot's context.

        .. versionadded:: 8.0
        """
        casemapping = {
            'ascii': identifiers.ascii_lower,
            'rfc1459': identifiers.rfc1459_lower,
            'rfc1459-strict': identifiers.rfc1459_strict_lower,
        }.get(self.isupport.get('CASEMAPPING'), identifiers.rfc1459_lower)
        chantypes = (
            self.isupport.get('CHANTYPES', identifiers.DEFAULT_CHANTYPES))

        return identifiers.Identifier(
            name,
            casemapping=casemapping,
            chantypes=chantypes,
        )

    def make_identifier_memory(self) -> memories.SopelIdentifierMemory:
        """Instantiate a SopelIdentifierMemory using the bot's context.

        This is a shortcut for :class:`~.memories.SopelIdentifierMemory`\'s most
        common use case, which requires remembering to pass the ``bot``\'s own
        :meth:`make_identifier` method so the ``SopelIdentifierMemory`` will
        cast its keys to :class:`~.tools.identifiers.Identifier`\\s that are
        compatible with what the bot tracks internally and sends with
        :class:`~.trigger.Trigger`\\s when a plugin callable runs.

        Calling this method is equivalent to the following::

            from sopel.tools import memories

            memories.SopelIdentifierMemory(
                identifier_factory=bot.make_identifier,
            )

        .. versionadded:: 8.0

        .. seealso::

            The :mod:`.tools.memories` module describes how to use
            :class:`~.tools.memories.SopelIdentifierMemory` and its siblings.

        """
        return memories.SopelIdentifierMemory(
            identifier_factory=self.make_identifier,
        )

    def safe_text_length(self, recipient: str) -> int:
        """Estimate a safe text length for an IRC message.

        :return: the maximum possible length of a message to ``recipient``

        When the bot sends a message to a recipient (channel or nick), it has
        512 bytes minus the command, arguments, various separators and trailing
        CRLF for its text. However, this is not what other users will see from
        the server; the message forwarded to other clients will be sent using
        this format::

            :nick!~user@hostname PRIVMSG #channel :text

        Which takes more bytes, reducing the maximum length available for a
        single line of text. This method computes a safe length of text that
        can be sent using ``PRIVMSG`` or ``NOTICE`` by subtracting the size
        required by the server to convey the bot's message.

        .. versionadded:: 8.0

        .. seealso::

            This method is useful when sending a message using :meth:`say`,
            and can be used with :func:`sopel.tools.get_sendable_message`.

        """
        # Clients "SHOULD" assume messages will be truncated at 512 bytes if
        # the LINELEN ISUPPORT token is not present.
        # See https://modern.ircdocs.horse/#linelen-parameter
        max_line_length = self.isupport.get('LINELEN', 512)

        if self.hostmask is not None:
            hostmask_length = len(self.hostmask)
        else:
            # calculate maximum possible length, given current nick/username
            hostmask_length = (
                len(self.nick)  # own nick length
                + 1  # (! separator)
                + 1  # (for the optional ~ in user)
                + min(  # own ident length, capped to ISUPPORT or RFC maximum
                    len(self.user),
                    getattr(self.isupport, 'USERLEN', 9))
                + 1  # (@ separator)
                + 63  # <hostname> has a maximum length of 63 characters.
            )

        return (
            max_line_length
            - 1  # leading colon
            - hostmask_length  # calculated/maximum length of own hostmask prefix
            - 1  # space between prefix & command
            - 7  # PRIVMSG command
            - 1  # space before recipient
            - len(recipient.encode('utf-8'))  # target channel/nick (can contain Unicode)
            - 2  # space after recipient, colon before text
            - 2  # trailing CRLF
        )

    # Connection

    def get_irc_backend(
        self,
        host: str,
        port: int,
        source_address: tuple[str, int] | None,
    ) -> AbstractIRCBackend:
        """Set up the IRC backend based on the bot's settings.

        :return: the initialized IRC backend object
        """
        timeout = int(self.settings.core.timeout)
        ping_interval = int(self.settings.core.timeout_ping_interval)
        return AsyncioBackend(
            self,
            # connection
            host=host,
            port=port,
            source_address=source_address,
            # timeout
            server_timeout=timeout,
            ping_interval=ping_interval,
            # ssl
            use_ssl=self.settings.core.use_ssl,
            certfile=self.settings.core.client_cert_file,
            keyfile=self.settings.core.client_cert_file,
            verify_ssl=self.settings.core.verify_ssl,
            ca_certs=self.settings.core.ca_certs,
            ssl_ciphers=self.settings.core.ssl_ciphers,
            ssl_minimum_version=self.settings.core.ssl_minimum_version,
        )

    def run(self, host: str, port: int = 6667) -> None:
        """Connect to IRC server and run the bot forever.

        :param host: the IRC server hostname
        :param port: the IRC server port
        """
        source_address = ((self.settings.core.bind_host, 0)
                          if self.settings.core.bind_host else None)

        self.backend = self.get_irc_backend(host, port, source_address)
        try:
            self.backend.run_forever()
        except KeyboardInterrupt:
            # raised only when the bot is not connected
            LOGGER.warning('Keyboard Interrupt')
            raise

    def on_connect(self) -> None:
        """Handle successful establishment of IRC connection."""
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        LOGGER.info('Connected, initiating setup sequence')

        # Request list of server capabilities. IRCv3 servers will respond with
        # CAP * LS (which we handle in coretasks). v2 servers will respond with
        # 421 Unknown command, which we'll ignore
        LOGGER.debug('Sending CAP request')
        self.backend.send_command('CAP', 'LS', '302')

        # authenticate account if needed
        if self.settings.core.auth_method == 'server':
            LOGGER.debug('Sending server auth')
            self.backend.send_pass(self.settings.core.auth_password)
        elif self.settings.core.server_auth_method == 'server':
            LOGGER.debug('Sending server auth')
            self.backend.send_pass(self.settings.core.server_auth_password)

        LOGGER.debug('Sending nick "%s"', self.nick)
        self.backend.send_nick(self.nick)
        LOGGER.debug('Sending user "%s" (name: "%s")', self.user, self.name)
        self.backend.send_user(self.user, '0', '*', self.name)

    def on_message(self, message: str) -> None:
        """Handle an incoming IRC message.

        :param message: the received raw IRC message
        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self.last_raw_line = message

        pretrigger = trigger.PreTrigger(
            self.nick,
            message,
            url_schemes=self.settings.core.auto_url_schemes,
            identifier_factory=self.make_identifier,
            statusmsg_prefixes=self.isupport.get('STATUSMSG'),
        )
        if all(
            not self.capabilities.is_enabled(cap)
            for cap in ['account-tag', 'extended-join']
        ):
            pretrigger.tags.pop('account', None)

        if pretrigger.event == 'PING':
            self.backend.send_pong(pretrigger.args[-1])
        elif pretrigger.event == 'ERROR':
            LOGGER.error("ERROR received from server: %s", pretrigger.args[-1])
            self.backend.on_irc_error(pretrigger)

        self.dispatch(pretrigger)

    def on_message_sent(self, raw: str) -> None:
        """Handle any message sent through the connection.

        :param raw: raw text message sent through the connection

        When a message is sent through the IRC connection, the bot will log
        the raw message. If necessary, it will also simulate the
        `echo-message`_ feature of IRCv3.

        .. _echo-message: https://ircv3.net/irc/#echo-message
        """
        # Log raw message
        self.log_raw(raw, '>>')

        # Simulate echo-message
        no_echo = not self.capabilities.is_enabled('echo-message')
        echoed = ['PRIVMSG', 'NOTICE']
        if no_echo and any(raw.upper().startswith(cmd) for cmd in echoed):
            # Use the hostmask we think the IRC server is using for us,
            # or something reasonable if that's not available
            host = 'localhost'
            if self.settings.core.bind_host:
                host = self.settings.core.bind_host
            else:
                try:
                    host = self.hostmask or host
                except KeyError:
                    pass  # we tried, and that's good enough

            pretrigger = trigger.PreTrigger(
                self.nick,
                ":{0}!{1}@{2} {3}".format(self.nick, self.user, host, raw),
                url_schemes=self.settings.core.auto_url_schemes,
                identifier_factory=self.make_identifier,
                statusmsg_prefixes=self.isupport.get('STATUSMSG'),
            )
            self.dispatch(pretrigger)

    @deprecated(
        'This method was used to log errors with asynchat; '
        'use logging.getLogger("sopel.exception") instead.',
        version='8.0',
        removed_in='9.0',
    )
    def on_error(self) -> None:
        """Handle any uncaptured error in the bot itself."""
        LOGGER.error('Fatal error in core, please review exceptions log.')

        err_log = logging.getLogger('sopel.exceptions')
        err_log.error(
            'Fatal error in core, bot.on_error() was called.\n'
            'Last Line:\n%s',
            self.last_raw_line,
        )
        err_log.exception('Fatal error traceback')
        err_log.error('----------------------------------------')

        if self.error_count > 10:
            # quit if too many errors
            dt_seconds: float = 0.0
            if self.last_error_timestamp is not None:
                dt = datetime.now(timezone.utc) - self.last_error_timestamp
                dt_seconds = dt.total_seconds()

            if dt_seconds < 5:
                LOGGER.error('Too many errors, can\'t continue')
                os._exit(1)
            # remove 1 error per full 5s that passed since last error
            self.error_count = int(max(0, self.error_count - dt_seconds // 5))

        self.last_error_timestamp = datetime.now(timezone.utc)
        self.error_count = self.error_count + 1

    def rebuild_nick(self) -> None:
        """Rebuild nick as a new identifier.

        This method exists to update the casemapping rules for the
        :class:`~sopel.tools.identifiers.Identifier` that represents the bot's
        nick, e.g. after ISUPPORT info is received.

        .. versionadded:: 8.0
        """
        self._nick = self.make_identifier(str(self._nick))

    def change_current_nick(self, new_nick: str) -> None:
        """Change the current nick without configuration modification.

        :param new_nick: new nick to be used by the bot

        .. versionadded:: 7.1
        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self._nick = self.make_identifier(new_nick)
        LOGGER.debug('Sending nick "%s"', self.nick)
        self.backend.send_nick(self.nick)

    def on_close(self) -> None:
        """Call shutdown methods."""
        self._connection_registered.clear()
        self._shutdown()

    def _shutdown(self) -> None:
        """Handle shutdown tasks.

        Must be overridden by subclasses to do anything useful.
        """

    # Features

    @abc.abstractmethod
    def dispatch(self, pretrigger: trigger.PreTrigger) -> None:
        """Handle running the appropriate callables for an incoming message.

        :param pretrigger: Sopel PreTrigger object

        .. important::
            This method **MUST** be implemented by concrete subclasses.
        """

    def log_raw(self, line: str, prefix: str) -> None:
        """Log raw line to the raw log.

        :param line: the raw line
        :param prefix: additional information to prepend to the log line

        The ``prefix`` is usually either ``>>`` for an outgoing ``line`` or
        ``<<`` for a received one.
        """
        if not self.settings.core.log_raw:
            return
        logger = logging.getLogger('sopel.raw')
        logger.info("%s\t%r", prefix, line)

    def write(self, args: Iterable[str], text: str | None = None) -> None:
        """Send a command to the server.

        :param args: an iterable of strings, which will be joined by spaces
        :param text: a string that will be prepended with a ``:`` and added to
                     the end of the command

        ``args`` is an iterable of strings, which are joined by spaces.
        ``text`` is treated as though it were the final item in ``args``, but
        is preceded by a ``:``. This is a special case which means that
        ``text``, unlike the items in ``args``, may contain spaces (though this
        constraint is not checked by ``write``).

        In other words, both ``sopel.write(('PRIVMSG',), 'Hello, world!')``
        and ``sopel.write(('PRIVMSG', ':Hello, world!'))`` will send
        ``PRIVMSG :Hello, world!`` to the server.

        Newlines and carriage returns (``'\\n'`` and ``'\\r'``) are removed
        before sending. Additionally, if the message (after joining) is longer
        than than 510 characters, any remaining characters will not be sent.

        .. seealso::

            The connection backend is responsible for formatting and sending
            the message through the IRC connection. See the
            :meth:`sopel.irc.abstract_backends.AbstractIRCBackend.send_command`
            method for more information.

        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self.backend.send_command(*args, text=text)

    # IRC Commands

    def action(self, text: str, dest: str) -> None:
        """Send a CTCP ACTION PRIVMSG to a user or channel.

        :param text: the text to send in the CTCP ACTION
        :param dest: the destination of the CTCP ACTION

        The same loop detection and length restrictions apply as with
        :func:`say`, though automatic message splitting is not available.
        """
        self.say('\001ACTION {}\001'.format(text), dest)

    def join(self, channel: str, password: str | None = None) -> None:
        """Join a ``channel``.

        :param channel: the channel to join
        :param password: an optional channel password

        If ``channel`` contains a space, and no ``password`` is given, the
        space is assumed to split the argument into the channel to join and its
        password. ``channel`` should not contain a space if ``password``
        is given.
        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self.backend.send_join(channel, password=password)

    def kick(
        self,
        nick: str,
        channel: str,
        text: str | None = None,
    ) -> None:
        """Kick a ``nick`` from a ``channel``.

        :param nick: nick to kick out of the ``channel``
        :param channel: channel to kick ``nick`` from
        :param text: optional text for the kick

        The bot must be an operator in the specified channel for this to work.

        .. versionadded:: 7.0
        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self.backend.send_kick(channel, nick, reason=text)

    def notice(self, text: str, dest: str) -> None:
        """Send an IRC NOTICE to a user or channel (``dest``).

        :param text: the text to send in the NOTICE
        :param dest: the destination of the NOTICE
        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self.backend.send_notice(dest, text)

    def part(self, channel: str, msg: str | None = None) -> None:
        """Leave a channel.

        :param channel: the channel to leave
        :param msg: the message to display when leaving a channel
        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self.backend.send_part(channel, reason=msg)

    def quit(self, message: str | None = None) -> None:
        """Disconnect from IRC and close the bot.

        :param message: optional QUIT message to send (e.g. "Bye!")
        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        self._connection_registered.clear()
        self.backend.send_quit(reason=message)
        self.hasquit = True
        # Wait for acknowledgment from the server. Per RFC 2812 it should be
        # an ERROR message, but many servers just close the connection.
        # Either way is fine by us. Closing the connection now would mean that
        # stuff in the buffers that has not yet been processed would never be
        # processed. It would also release the main thread, which is
        # problematic because whomever called quit might still want to do
        # something before the main thread quits.

    def restart(self, message: str | None = None) -> None:
        """Disconnect from IRC and restart the bot.

        :param message: optional QUIT message to send (e.g. "Be right back!")
        """
        self.wantsrestart = True
        self.quit(message)

    def reply(
        self,
        text: str,
        dest: str,
        reply_to: str,
        notice: bool = False,
    ) -> None:
        """Send a PRIVMSG to a user or channel, prepended with ``reply_to``.

        :param text: the text of the reply
        :param dest: the destination of the reply
        :param reply_to: the nickname that the reply will be prepended with
        :param notice: whether to send the reply as a ``NOTICE`` or not,
                       defaults to ``False``

        If ``notice`` is ``True``, send a ``NOTICE`` rather than a ``PRIVMSG``.

        The same loop detection and length restrictions apply as with
        :meth:`say`, though automatic message splitting is not available.
        """
        text = '%s: %s' % (reply_to, text)
        if notice:
            self.notice(text, dest)
        else:
            self.say(text, dest)

    def say(
        self,
        text: str,
        recipient: str,
        max_messages: int = 1,
        truncation: str = '',
        trailing: str = '',
    ) -> None:
        """Send a ``PRIVMSG`` to a user or channel.

        :param text: the text to send
        :param recipient: the message recipient
        :param max_messages: split ``text`` into at most this many messages
                             if it is too long to fit in one (optional)
        :param truncation: string to append if ``text`` is too long to fit in
                           a single message, or into the last message if
                           ``max_messages`` is greater than 1 (optional)
        :param trailing: string to append after ``text`` and (if used)
                         ``truncation`` (optional)

        By default, this will attempt to send the entire ``text`` in one
        message. If the text is too long for the server, it may be truncated.

        If ``max_messages`` is given, the ``text`` will be split into at most
        that many messages. The split is made at the last space character
        before the "safe length" (which is calculated based on the bot's
        nickname and hostmask), or exactly at the "safe length" if no such
        space character exists.

        If the ``text`` is too long to fit into the specified number of messages
        using the above splitting, the final message will contain the entire
        remainder, which may be truncated by the server. You can specify
        ``truncation`` to tell Sopel how it should indicate that the remaining
        ``text`` was cut off. Note that the ``truncation`` parameter must
        include leading whitespace if you desire any between it and the
        truncated text.

        The ``trailing`` parameter is *always* appended to ``text``, after the
        point where ``truncation`` would be inserted if necessary. It's useful
        for making sure e.g. a link is always included, even if the summary your
        plugin fetches is too long to fit.

        Here are some examples of how the ``truncation`` and ``trailing``
        parameters work, using an artificially low maximum line length::

            # bot.say() outputs <text> + <truncation?> + <trailing>
            #                   always     if needed       always

            bot.say(
                '"This is a short quote.',
                truncation=' […]',
                trailing='"')
            # Sopel says: "This is a short quote."

            bot.say(
                '"This quote is very long and will not fit on a line.',
                truncation=' […]',
                trailing='"')
            # Sopel says: "This quote is very long […]"

            bot.say(
                # note the " included at the end this time
                '"This quote is very long and will not fit on a line."',
                truncation=' […]')
            # Sopel says: "This quote is very long […]
            # The ending " goes missing

        .. versionadded:: 7.1

            The ``truncation`` and ``trailing`` parameters.

        """
        if self.backend is None:
            raise RuntimeError(ERR_BACKEND_NOT_INITIALIZED)

        excess = ''

        if not isinstance(text, str):
            # Make sure we are dealing with a Unicode string
            text = text.decode('utf-8')

        safe_length = self.safe_text_length(recipient)
        if trailing and max_messages == 1:
            # last message needs to leave room for `trailing`
            safe_length -= len(trailing.encode('utf-8'))

        # only think about `truncation` if we need to
        if safe_length < len(text.encode('utf-8')):
            if max_messages == 1:
                # last message needs to leave room for `truncation`
                # if it's still too long to fit in the line
                safe_length -= len(truncation.encode('utf-8'))
            text, excess = tools.get_sendable_message(text, safe_length)
            if max_messages == 1:
                text += truncation

        if max_messages == 1:
            # ALWAYS append `trailing` to the last message;
            # its size is included in the initial `safe_length` check
            text += trailing

        flood_max_wait = self.settings.core.flood_max_wait
        flood_burst_lines = self.settings.core.flood_burst_lines
        flood_refill_rate = self.settings.core.flood_refill_rate
        flood_empty_wait = self.settings.core.flood_empty_wait

        flood_text_length = self.settings.core.flood_text_length
        flood_penalty_ratio = self.settings.core.flood_penalty_ratio

        antiloop_threshold = min(10, self.settings.core.antiloop_threshold)
        antiloop_window = self.settings.core.antiloop_window
        antiloop_repeat_text = self.settings.core.antiloop_repeat_text
        antiloop_silent_after = self.settings.core.antiloop_silent_after

        with self.sending:
            recipient_id = self.make_identifier(recipient)
            recipient_stack = self.stack.setdefault(recipient_id, {
                'messages': deque(maxlen=10),
                'flood_left': flood_burst_lines,
            })

            if recipient_stack['messages']:
                elapsed = time.time() - recipient_stack['messages'][-1][0]
            else:
                # Default to a high enough value that we won't care.
                # Five minutes should be enough not to matter anywhere below.
                elapsed = 300

            # If flood bucket is empty, refill the appropriate number of lines
            # based on how long it's been since our last message to recipient
            if not recipient_stack['flood_left']:
                recipient_stack['flood_left'] = min(
                    flood_burst_lines,
                    int(elapsed) * flood_refill_rate)

            # If it's too soon to send another message, wait
            if not recipient_stack['flood_left']:
                penalty = 0

                if flood_penalty_ratio > 0:
                    penalty_ratio = flood_text_length * flood_penalty_ratio
                    text_length_overflow = float(
                        max(0, len(text) - flood_text_length))
                    penalty = text_length_overflow / penalty_ratio

                # Maximum wait time is 2 sec by default
                initial_wait_time = flood_empty_wait + penalty
                wait = min(initial_wait_time, flood_max_wait)
                if elapsed < wait:
                    sleep_time = wait - elapsed
                    LOGGER.debug(
                        'Flood protection wait time: %.3fs; '
                        'elapsed time: %.3fs; '
                        'initial wait time (limited to %.3fs): %.3fs '
                        '(including %.3fs of penalty).',
                        sleep_time,
                        elapsed,
                        flood_max_wait,
                        initial_wait_time,
                        penalty,
                    )
                    time.sleep(sleep_time)

            # Loop detection
            if antiloop_threshold > 0 and elapsed < antiloop_window:
                messages = [m[1] for m in recipient_stack['messages']]

                # If what we're about to send repeated at least N times
                # in the anti-looping window, replace it
                if messages.count(text) >= antiloop_threshold:
                    text = antiloop_repeat_text
                    if messages.count(text) >= antiloop_silent_after:
                        # If we've already said that N times, discard message
                        return

            self.backend.send_privmsg(recipient, text)

            # update recipient metadata
            flood_left = recipient_stack['flood_left'] - 1
            recipient_stack['flood_left'] = max(0, flood_left)
            recipient_stack['messages'].append((time.time(), text))

        # Now that we've sent the first part, we need to send the rest if
        # requested. Doing so recursively seems simpler than iteratively.
        if max_messages > 1 and excess:
            self.say(excess, recipient, max_messages - 1, truncation, trailing)



"""Types for creating section definitions.

A section definition consists of a subclass of :class:`StaticSection`, on which
any number of subclasses of :class:`BaseValidated` (a few common ones of which
are available in this module) are assigned as attributes. These descriptors
define how to read values from, and write values to, the config file.

As an example, if one wanted to define the ``[spam]`` section as having an
``eggs`` option, which contains a list of values, they could do this:

    >>> class SpamSection(StaticSection):
    ...     eggs = ListAttribute('eggs')
    ...
    >>> SpamSection(config, 'spam')
    >>> print(config.spam.eggs)
    []
    >>> config.spam.eggs = ['goose', 'turkey', 'duck', 'chicken', 'quail']
    >>> print(config.spam.eggs)
    ['goose', 'turkey', 'duck', 'chicken', 'quail']
    >>> config.spam.eggs = 'herring'
    Traceback (most recent call last):
        ...
    ValueError: ListAttribute value must be a list.
"""

from __future__ import annotations

import abc
import getpass
import logging
import os.path
import re

from sopel.lifecycle import deprecated


LOGGER = logging.getLogger(__name__)


class NO_DEFAULT:
    """A special value to indicate that there should be no default."""


class StaticSection:
    """A configuration section with parsed and validated settings.

    This class is intended to be subclassed and customized with added
    attributes containing :class:`BaseValidated`-based objects.

    .. note::

        By convention, subclasses of ``StaticSection`` are named with the
        plugin's name in CamelCase, plus the suffix ``Section``. For example, a
        plugin named ``editor`` might name its subclass ``EditorSection``; a
        ``do_stuff`` plugin might name its subclass ``DoStuffSection`` (its
        name converted from ``snake_case`` to ``CamelCase``).

        However, this is *only* a convention. Any class name that is legal in
        Python will work just fine.

    """
    def __init__(self, config, section_name, validate=True):
        if not config.parser.has_section(section_name):
            config.parser.add_section(section_name)
        self._parent = config
        self._parser = config.parser
        self._section_name = section_name

        for value in dir(self):
            if value in ('_parent', '_parser', '_section_name'):
                # ignore internal attributes
                continue

            try:
                getattr(self, value)
            except ValueError as e:
                raise ValueError(
                    'Invalid value for {}.{}: {}'.format(
                        section_name, value, str(e)))
            except AttributeError:
                if validate:
                    raise ValueError(
                        'Missing required value for {}.{}'.format(
                            section_name, value))

    def configure_setting(self, name, prompt, default=NO_DEFAULT):
        """Return a validated value for this attribute from the terminal.

        :param str name: the name of the attribute to configure
        :param str prompt: the prompt text to display in the terminal
        :param default: the value to be used if the user does not enter one
        :type default: depends on subclass

        If ``default`` is passed, it will be used if no value is given by the
        user. If it is not passed, the current value of the setting, or the
        default value if it's unset, will be used. Note that if ``default`` is
        passed, the current value of the setting will be ignored, even if it is
        not the attribute's default.
        """
        attribute = getattr(self.__class__, name)
        if default is NO_DEFAULT and not attribute.is_secret:
            try:
                # get current value of this setting to use as prompt default
                default = getattr(self, name)
            except AttributeError:
                # there is no current value; that's OK
                pass
            except ValueError:
                print('The configured value for this option was invalid.')
                if attribute.default is not NO_DEFAULT:
                    default = attribute.default
        while True:
            try:
                value = attribute.configure(
                    prompt, default, self._parent, self._section_name)
            except ValueError as exc:
                print(exc)
            else:
                break
        setattr(self, name, value)


class BaseValidated(abc.ABC):
    """The base type for a setting descriptor in a :class:`StaticSection`.

    :param str name: the attribute name to use in the config file
    :param default: the value to be returned if the setting has no value
                    (optional; defaults to :obj:`None`)
    :type default: mixed
    :param bool is_secret: tell if the attribute is secret/a password
                           (optional; defaults to ``False``)

    ``default`` also can be set to :const:`sopel.config.types.NO_DEFAULT`, if
    the value *must* be configured by the user (i.e. there is no suitable
    default value). Trying to read an empty ``NO_DEFAULT`` value will raise
    :class:`AttributeError`.

    .. important::

        Setting names SHOULD follow *snake_case* naming rules:

        * use only lowercase letters, digits, and underscore (``_``)
        * SHOULD NOT start with a digit

        Deviations from *snake_case* can break the following operations:

        * :ref:`accessing the setting <sopel.config>` from Python code using
          the :class:`~.Config` object's attributes
        * :ref:`overriding the setting's value <Overriding individual
          settings>` using environment variables

    """
    def __init__(self, name, default=None, is_secret=False):
        self.name = name
        """Name of the attribute."""
        self.default = default
        """Default value for this attribute.

        If not specified, the attribute's default value will be ``None``.
        """
        self.is_secret = bool(is_secret)
        """Tell if the attribute is secret/a password.

        The default value is ``False`` (not secret).

        Sopel's configuration can contain passwords, secret keys, and other
        private information that must be treated as sensitive data. Such
        options should be marked as "secret" with this attribute.
        """

    def configure(self, prompt, default, parent, section_name):
        """Parse and return a value from user's input.

        :param str prompt: text to show the user
        :param mixed default: default value used if no input given
        :param parent: usually the parent Config object
        :type parent: :class:`~sopel.config.Config`
        :param str section_name: the name of the containing section

        This method displays the ``prompt`` and waits for user's input on the
        terminal. If no input is provided (i.e. the user just presses "Enter"),
        the ``default`` value is returned instead.

        If :attr:`.is_secret` is ``True``, the input will be hidden, using the
        built-in :func:`~getpass.getpass` function.
        """
        if default is not NO_DEFAULT and default is not None:
            prompt = '{} [{}]'.format(prompt, default)

        if self.is_secret:
            value = getpass.getpass(prompt + ' (hidden input) ')
        else:
            value = input(prompt + ' ')

        if not value and default is NO_DEFAULT:
            raise ValueError("You must provide a value for this option.")

        value = value or default
        section = getattr(parent, section_name)

        return self._parse(value, parent, section)

    @abc.abstractmethod
    def serialize(self, value, *args, **kwargs):
        """Take some object, and return the string to be saved to the file."""

    @abc.abstractmethod
    def parse(self, value, *args, **kwargs):
        """Take a string from the file, and return the appropriate object."""

    def __get__(self, instance, owner=None):
        if instance is None:
            # If instance is None, we're getting from a section class, not an
            # instance of a section class. It makes the wizard code simpler
            # (and is really just more intuitive) to return the descriptor
            # instance here.
            return self

        value = None
        env_name = 'SOPEL_%s_%s' % (instance._section_name.upper(), self.name.upper())
        if env_name in os.environ:
            value = os.environ.get(env_name)
        elif instance._parser.has_option(instance._section_name, self.name):
            value = instance._parser.get(instance._section_name, self.name)

        settings = instance._parent
        section = getattr(settings, instance._section_name)
        return self._parse(value, settings, section)

    def _parse(self, value, settings, section):
        if value is not None:
            return self.parse(value)
        if self.default is not NO_DEFAULT:
            return self.default
        raise AttributeError(
            "Missing required value for {}.{}".format(
                section._section_name, self.name
            )
        )

    def __set__(self, instance, value):
        if value is None:
            if self.default == NO_DEFAULT:
                raise ValueError('Cannot unset an option with a required value.')
            instance._parser.remove_option(instance._section_name, self.name)
            return

        settings = instance._parent
        section = getattr(settings, instance._section_name)
        value = self._serialize(value, settings, section)
        instance._parser.set(instance._section_name, self.name, value)

    def _serialize(self, value, settings, section):
        return self.serialize(value)

    def __delete__(self, instance):
        instance._parser.remove_option(instance._section_name, self.name)


def _parse_boolean(value):
    if value is True or value == 1:
        return value
    if isinstance(value, str):
        return value.lower() in ['1', 'yes', 'y', 'true', 'on']
    return bool(value)


def _serialize_boolean(value):
    return 'true' if _parse_boolean(value) else 'false'


@deprecated(
    reason='Use BooleanAttribute instead of ValidatedAttribute with parse=bool',
    version='7.1',
    warning_in='8.0',
    removed_in='9.0',
    stack_frame=-2,
)
def _deprecated_special_bool_handling(serialize):
    if not serialize or serialize == bool:
        serialize = _serialize_boolean

    return _parse_boolean, serialize


class ValidatedAttribute(BaseValidated):
    """A descriptor for settings in a :class:`StaticSection`.

    :param str name: the attribute name to use in the config file
    :param parse: a function to be used to read the string and create the
                  appropriate object (optional; the string value will be
                  returned as-is if not set)
    :type parse: :term:`function`
    :param serialize: a function that, given an object, should return a string
                      that can be written to the config file safely (optional;
                      defaults to :class:`str`)
    :type serialize: :term:`function`
    :param bool is_secret: ``True`` when the attribute should be considered
                           a secret, like a password (default to ``False``)
    """
    def __init__(self,
                 name,
                 parse=None,
                 serialize=None,
                 default=None,
                 is_secret=False):
        super().__init__(name, default=default, is_secret=is_secret)

        if parse == bool:
            parse, serialize = _deprecated_special_bool_handling(serialize)

        # ignore typing errors on these monkeypatches for now
        # TODO: more dedicated subtypes; deprecate `parse`/`serialize` args
        self.parse = parse or self.parse  # type: ignore
        self.serialize = serialize or self.serialize  # type: ignore

    def serialize(self, value):
        """Return the ``value`` as a Unicode string.

        :param value: the option value
        :rtype: str
        """
        return str(value)

    def parse(self, value):
        """No-op: simply returns the given ``value``, unchanged.

        :param str value: the string read from the config file
        :rtype: str
        """
        return value

    def configure(self, prompt, default, parent, section_name):
        if self.parse == _parse_boolean:
            prompt += ' (y/n)'
            default = 'y' if default else 'n'
        return super().configure(prompt, default, parent, section_name)


class BooleanAttribute(BaseValidated):
    """A descriptor for Boolean settings in a :class:`StaticSection`.

    :param str name: the attribute name to use in the config file
    :param bool default: the default value to use if this setting is not
                         present in the config file

    If the ``default`` value is not specified, it will be ``False``.
    """
    def __init__(self, name, default=False):
        super().__init__(name, default=default, is_secret=False)

    def configure(self, prompt, default, parent, section_name):
        """Parse and return a value from user's input.

        :param str prompt: text to show the user
        :param bool default: default value used if no input given
        :param parent: usually the parent Config object
        :type parent: :class:`~sopel.config.Config`
        :param str section_name: the name of the containing section

        This method displays the ``prompt`` and waits for user's input on the
        terminal. If no input is provided (i.e. the user just presses "Enter"),
        the ``default`` value is returned instead.
        """
        prompt = '{} ({})'.format(prompt, 'Y/n' if default else 'y/N')
        value = input(prompt + ' ') or default
        section = getattr(parent, section_name)
        return self._parse(value, parent, section)

    def serialize(self, value):
        """Convert a Boolean value to a string for saving to the config file.

        :param bool value: the value to serialize
        """
        return 'true' if self.parse(value) else 'false'

    def parse(self, value):
        """Parse a limited set of values/objects into Boolean representations.

        :param mixed value: the value to parse

        The literal values ``True`` or ``1`` will be parsed as ``True``. The
        strings ``'1'``, ``'yes'``, ``'y'``, ``'true'``, ``'enable'``,
        ``'enabled'``, and ``'on'`` will also be parsed as ``True``,
        regardless of case. All other values will be parsed as ``False``.
        """
        if value is True or value == 1:
            return True
        if isinstance(value, str):
            return value.lower() in [
                '1',
                'enable',
                'enabled',
                'on',
                'true',
                'y',
                'yes',
            ]
        return bool(value)

    def __set__(self, instance, value):
        if value is None:
            instance._parser.remove_option(instance._section_name, self.name)
            return

        settings = instance._parent
        section = getattr(settings, instance._section_name)
        value = self._serialize(value, settings, section)
        instance._parser.set(instance._section_name, self.name, value)


class SecretAttribute(ValidatedAttribute):
    """A config attribute containing a value which must be kept secret.

    This attribute is always considered to be secret/sensitive data, but
    otherwise behaves like other any option.
    """
    def __init__(self, name, parse=None, serialize=None, default=None):
        super().__init__(
            name,
            parse=parse,
            serialize=serialize,
            default=default,
            is_secret=True,
        )


class ListAttribute(BaseValidated):
    """A config attribute containing a list of string values.

    :param str name: the attribute name to use in the config file
    :param strip: whether to strip whitespace from around each value
                  (optional, deprecated; applies only to legacy comma-separated
                  lists; multi-line lists are always stripped)
    :type strip: bool
    :param default: the default value if the config file does not define a
                    value for this option; to require explicit configuration,
                    use :const:`sopel.config.types.NO_DEFAULT` (optional)
    :type default: list

    From this :class:`StaticSection`::

        class SpamSection(StaticSection):
            cheeses = ListAttribute('cheeses')

    the option will be exposed as a Python :class:`list`::

        >>> config.spam.cheeses
        ['camembert', 'cheddar', 'reblochon', '#brie']

    which comes from this configuration file:

    .. code-block:: ini

        [spam]
        cheeses =
            camembert
            cheddar
            reblochon
            "#brie"

    Note that the ``#brie`` item starts with a ``#``, hence the double quote:
    without these quotation marks, the config parser would think it's a
    comment. The quote/unquote is managed automatically by this field, and
    if and only if it's necessary (see :meth:`parse` and :meth:`serialize`).

    .. versionchanged:: 7.0

        The option's value will be split on newlines by default. In this
        case, the ``strip`` parameter has no effect.

        See the :meth:`parse` method for more information.

    .. note::

        **About:** backward compatibility with comma-separated values.

        A :class:`ListAttribute` option used to allow to write, on a single
        line, the values separated by commas. It is still technically possible
        while raising a deprecation warning.

        In Sopel 7.x this behavior was discouraged; as of Sopel 8.x it is now
        deprecated with warnings, and it will be removed in Sopel 9.x.

        Bot owners should update their configurations to use newlines instead
        of commas.

        The comma delimiter fallback does not support commas within items in
        the list.
    """
    DELIMITER = ','  # Deprecated, will be removed in Sopel 9
    QUOTE_REGEX = re.compile(r'^"(?P<value>#.*)"$')
    """Regex pattern to match value that requires quotation marks.

    This pattern matches values that start with ``#`` inside quotation marks
    only: ``"#sopel"`` will match, but ``"sopel"`` won't, and neither will any
    variant that doesn't conform to this pattern.
    """

    def __init__(self, name, strip=True, default=None):
        default = default or []
        super().__init__(name, default=default)
        self.strip = strip  # Warn in Sopel 9.x and remove in Sopel 10.x

    def parse(self, value):
        """Parse ``value`` into a list.

        :param str value: a multi-line string of values to parse into a list
        :return: a list of items from ``value``
        :rtype: list

        .. versionchanged:: 7.0

            The value is now split on newlines, with fallback to comma
            when there is no newline in ``value``.

            When modified and saved to a file, items will be stored as a
            multi-line string (see :meth:`serialize`).

        .. versionchanged:: 8.0

            When the value contains a delimiter without newline, it warns the
            user to switch to a multi-line value, without a delimiter.

        """
        if "\n" in value:
            items = (
                # remove trailing comma
                # because `value,\nother` is valid in Sopel 7.x
                item.strip(self.DELIMITER).strip()
                for item in value.splitlines())
        else:
            # this behavior was discouraged in Sopel 7.x (in the documentation)
            # this behavior is now deprecated in Sopel 8.x
            # this behavior will be removed from Sopel 9.x
            items = value.split(self.DELIMITER)
            if self.DELIMITER in value:
                # trigger for "one, two" and "first line,"
                # a single line without the delimiter is fine
                LOGGER.warning(
                    'Using "%s" as item delimiter in option "%s" '
                    'is deprecated and will be removed in Sopel 9; '
                    'use multi-line instead',
                    self.DELIMITER, self.name)

        items = (self.parse_item(item) for item in items if item)
        if self.strip:
            return [item.strip() for item in items]

        return list(items)

    def parse_item(self, item):
        """Parse one ``item`` from the list.

        :param str item: one item from the list to parse
        :rtype: str

        If ``item`` matches the :attr:`QUOTE_REGEX` pattern, then it will be
        unquoted. Otherwise it's returned as-is.
        """
        result = self.QUOTE_REGEX.match(item)
        if result:
            return result.group('value')
        return item

    def serialize(self, value):
        """Serialize ``value`` into a multi-line string.

        :param list value: the input list
        :rtype: str
        :raise ValueError: if ``value`` is the wrong type (i.e. not a list)
        """
        if not isinstance(value, (list, set)):
            raise ValueError('ListAttribute value must be a list.')
        elif not value:
            # return an empty string when there is no value
            return ''

        # we ensure to read a newline, even with only one value in the list
        # this way, comma will be ignored when the configuration file
        # is read again later
        return '\n' + '\n'.join(self.serialize_item(item) for item in value)

    def serialize_item(self, item):
        """Serialize an ``item`` from the list value.

        :param str item: one item of the list to serialize
        :rtype: str

        If ``item`` starts with a ``#`` it will be quoted in order to prevent
        the config parser from thinking it's a comment.
        """
        if item.startswith('#'):
            # we need to protect item that would otherwise appear as comment
            return '"%s"' % item
        return item

    def configure(self, prompt, default, parent, section_name):
        each_prompt = '?'
        if isinstance(prompt, tuple):
            each_prompt = prompt[1]
            prompt = prompt[0]

        if default is not NO_DEFAULT:
            default_prompt = ','.join(['"{}"'.format(item) for item in default])
            prompt = '{} [{}]'.format(prompt, default_prompt)
        else:
            default = []
        print(prompt)
        values: list[str] = []
        value = input(each_prompt + ' ') or default
        if (value == default) and not default:
            value = ''
        while value:
            if value == default:
                values.extend(value)
            else:
                values.append(value)
            value = input(each_prompt + ' ')

        section = getattr(parent, section_name)
        values = self._serialize(values, parent, section)
        return self._parse(values, parent, section)


class ChoiceAttribute(BaseValidated):
    """A config attribute which must be one of a set group of options.

    :param str name: the attribute name to use in the config file
    :param choices: acceptable values; currently, only strings are supported
    :type choices: list or tuple
    :param default: which choice to use if none is set in the config file; to
                    require explicit configuration, use
                    :const:`sopel.config.types.NO_DEFAULT` (optional)
    :type default: str
    """
    def __init__(self, name, choices, default=None):
        super().__init__(name, default=default)
        self.choices = choices

    def parse(self, value):
        """Check the loaded ``value`` against the valid ``choices``.

        :param str value: the value loaded from the config file
        :return: the ``value``, if it is valid
        :rtype: str
        :raise ValueError: if ``value`` is not one of the valid ``choices``
        """
        if value in self.choices:
            return value
        else:
            raise ValueError(
                '{!r} is not one of the valid choices: {}'
                .format(value, ', '.join(self.choices))
            )

    def serialize(self, value):
        """Make sure ``value`` is valid and safe to write in the config file.

        :param str value: the value needing to be saved
        :return: the ``value``, if it is valid
        :rtype: str
        :raise ValueError: if ``value`` is not one of the valid ``choices``
        """
        if value in self.choices:
            return value
        else:
            raise ValueError(
                '{!r} is not one of the valid choices: {}'
                .format(value, ', '.join(self.choices))
            )


class FilenameAttribute(BaseValidated):
    """A config attribute which must be a file or directory.

    :param str name: the attribute name to use in the config file
    :param relative: whether the path should be relative to the location of
                     the config file (optional; note that absolute paths will
                     always be interpreted as absolute)
    :type relative: bool
    :param directory: whether the path should indicate a directory, rather
                      than a file (optional)
    :type directory: bool
    :param default: the value to use if none is defined in the config file; to
                    require explicit configuration, use
                    :const:`sopel.config.types.NO_DEFAULT` (optional)
    :type default: str
    """
    def __init__(self, name, relative=True, directory=False, default=None):
        super().__init__(name, default=default)
        self.relative = relative
        self.directory = directory

    def _parse(self, value, settings, section):
        if value is None:
            if self.default == NO_DEFAULT:
                raise AttributeError(
                    "Missing required value for {}.{}".format(
                        section._section_name, self.name
                    )
                )
            value = self.default

        if not value:
            return self.parse(value)

        if value.startswith('"') and value.endswith('"'):
            value = value.strip('"')
        elif value.startswith("'") and value.endswith("'"):
            value = value.strip("'")

        result = os.path.expanduser(value)
        if not os.path.isabs(result):
            if not self.relative:
                raise ValueError("Value must be an absolute path.")
            result = os.path.join(settings.homedir, result)

        return self.parse(result)

    def _serialize(self, value, settings, section):
        """Used to validate ``value`` when it is changed at runtime.

        :param settings: the config object which contains this attribute
        :type settings: :class:`~sopel.config.Config`
        :param section: the config section which contains this attribute
        :type section: :class:`~StaticSection`
        :return: the ``value``, if it is valid
        :rtype: str
        :raise ValueError: if the ``value`` is not valid
        """
        self._parse(value, settings, section)
        return self.serialize(value)

    def parse(self, value):
        """Parse ``value`` as a path on the filesystem to check.

        :param str value: the path to check
        :rtype: str
        :raise ValueError: if the directory or file doesn't exist and cannot
                           be created

        If there is no ``value``, then this returns ``None``. Otherwise, it'll
        check if the directory or file exists. If it doesn't, it'll try to
        create it.
        """
        if not value:
            return None

        if self.directory and not os.path.isdir(value):
            try:
                os.makedirs(value)
            except (IOError, OSError):
                raise ValueError(
                    "Value must be an existing or creatable directory.")
        if not self.directory and not os.path.isfile(value):
            try:
                with open(value, 'w') as f:
                    f.close()
            except (IOError, OSError):
                raise ValueError("Value must be an existing or creatable file.")
        return value

    def serialize(self, value):
        """Directly return the ``value`` without modification.

        :param str value: the value needing to be saved
        :return: the unaltered ``value``, if it is valid
        :rtype: str

        Managing the filename is done by other methods (:meth:`parse`), and
        this method is a no-op: this way it ensures that relative paths won't
        be replaced by absolute ones.
        """
        return value  # So that it's still relative



"""Sopel's plugin handlers.

.. versionadded:: 7.0

Between a plugin (or "module") and Sopel's core, Plugin Handlers are used. It
is an interface (defined by the :class:`AbstractPluginHandler` abstract class),
that acts as a proxy between Sopel and the plugin, making a clear separation
between how the bot behaves and how the plugins work.

From the :class:`~sopel.bot.Sopel` class, a plugin must be:

* loaded, using :meth:`~AbstractPluginHandler.load`
* setup (if required), using :meth:`~AbstractPluginHandler.setup`
* and eventually registered using :meth:`~AbstractPluginHandler.register`

Each subclass of :class:`AbstractPluginHandler` must implement its methods in
order to be used in the application.

At the moment, three types of plugin are handled:

* :class:`PyModulePlugin`: manages plugins that can be imported as Python
  module from a Python package, i.e. where ``from package import name`` works
* :class:`PyFilePlugin`: manages plugins that are Python files on the filesystem
  or Python directory (with an ``__init__.py`` file inside), that cannot be
  directly imported and extra steps are necessary
* :class:`EntryPointPlugin`: manages plugins that are declared by an entry
  point; it otherwise behaves like a :class:`PyModulePlugin`

All expose the same interface and thereby abstract the internal implementation
away from the rest of the application.

.. important::

    This is all relatively new. Its usage and documentation is for Sopel core
    development and advanced developers. It is subject to rapid changes
    between versions without much (or any) warning.

    Do **not** build your plugin based on what is here, you do **not** need to.

"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import annotations

import abc
import importlib
import importlib.util
import inspect
import itertools
import logging
import os
import sys
from typing import Optional, TYPE_CHECKING, TypedDict

from sopel import __version__ as release, loader, plugin as plugin_decorators
from . import exceptions


if TYPE_CHECKING:
    from sopel.bot import Sopel
    from types import ModuleType


LOGGER = logging.getLogger(__name__)


class PluginMetaDescription(TypedDict):
    """Meta description of a plugin, as a dictionary.

    This dictionary is expected to contain specific keys:

    * name: a short name for the plugin
    * label: a descriptive label for the plugin; see
      :meth:`~sopel.plugins.handlers.AbstractPluginHandler.get_label`
    * type: the plugin's type
    * source: the plugin's source
      (filesystem path, python module/import path, etc.)
    * version: the plugin's version string if available, otherwise ``None``
    """
    name: str
    label: str
    type: str
    source: str
    version: Optional[str]


class AbstractPluginHandler(abc.ABC):
    """Base class for plugin handlers.

    This abstract class defines the interface Sopel uses to
    configure, load, shutdown, etc. a Sopel plugin (or "module").

    It is through this interface that Sopel will interact with its plugins,
    whether internal (from ``sopel.builtins``) or external (from the Python
    files in a directory, to ``sopel_modules.*`` subpackages).

    Sopel's loader will create a "Plugin Handler" for each plugin it finds, to
    which it then delegates loading the plugin, listing its functions
    (commands, jobs, etc.), configuring it, and running any required actions
    on shutdown (either upon exiting Sopel or unloading that plugin).
    """

    name: str
    """Plugin identifier.

    The name of a plugin identifies this plugin: when Sopel loads a plugin,
    it will store its information under that identifier.
    """

    @abc.abstractmethod
    def load(self):
        """Load the plugin.

        This method must be called first, in order to setup, register, shutdown,
        or configure the plugin later.
        """

    @abc.abstractmethod
    def reload(self):
        """Reload the plugin.

        This method can be called once the plugin is already loaded. It will
        take care of reloading the plugin from its source.
        """

    @abc.abstractmethod
    def get_label(self) -> str:
        """Retrieve a display label for the plugin.

        :return: a human readable label for display purpose
        :rtype: str

        This method should, at least, return ``<module_name> plugin``.
        """

    @abc.abstractmethod
    def get_meta_description(self) -> PluginMetaDescription:
        """Retrieve a meta description for the plugin.

        :return: Metadata about the plugin
        :rtype: :class:`dict`

        The expected keys are detailed in :class:`PluginMetaDescription`.
        """

    @abc.abstractmethod
    def get_version(self):
        """Retrieve the plugin's version.

        :return: the plugin's version string
        :rtype: str
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_loaded(self) -> bool:
        """Tell if the plugin is loaded or not.

        :return: ``True`` if the plugin is loaded, ``False`` otherwise
        :rtype: bool

        This must return ``True`` if the :meth:`load` method has been called
        with success.
        """

    @abc.abstractmethod
    def setup(self, bot):
        """Run the plugin's setup action.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """

    @abc.abstractmethod
    def has_setup(self) -> bool:
        """Tell if the plugin has a setup action.

        :return: ``True`` if the plugin has a setup, ``False`` otherwise
        :rtype: bool
        """

    @abc.abstractmethod
    def get_capability_requests(self) -> list[plugin_decorators.capability]:
        """Retrieve the plugin's list of capability requests."""

    @abc.abstractmethod
    def register(self, bot):
        """Register the plugin with the ``bot``.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """

    @abc.abstractmethod
    def unregister(self, bot):
        """Unregister the plugin from the ``bot``.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """

    @abc.abstractmethod
    def shutdown(self, bot):
        """Run the plugin's shutdown action.

        :param bot: instance of Sopel
        :type bot: :class:`sopel.bot.Sopel`
        """

    @abc.abstractmethod
    def has_shutdown(self) -> bool:
        """Tell if the plugin has a shutdown action.

        :return: ``True`` if the plugin has a ``shutdown`` action, ``False``
                 otherwise
        :rtype: bool
        """

    @abc.abstractmethod
    def configure(self, settings):
        """Configure Sopel's ``settings`` for this plugin.

        :param settings: Sopel's configuration
        :type settings: :class:`sopel.config.Config`

        This method will be called by Sopel's configuration wizard.
        """

    @abc.abstractmethod
    def has_configure(self) -> bool:
        """Tell if the plugin has a configure action.

        :return: ``True`` if the plugin has a ``configure`` action, ``False``
                 otherwise
        :rtype: bool
        """


class PyModulePlugin(AbstractPluginHandler):
    """Sopel plugin loaded from a Python module or package.

    A :class:`PyModulePlugin` represents a Sopel plugin that is a Python
    module (or package) that can be imported directly.

    This::

        >>> import sys
        >>> from sopel.plugins.handlers import PyModulePlugin
        >>> plugin = PyModulePlugin('xkcd', 'sopel.builtins')
        >>> plugin.module_name
        'sopel.builtins.xkcd'
        >>> plugin.load()
        >>> plugin.module_name in sys.modules
        True

    Is the same as this::

        >>> import sys
        >>> from sopel.builtins import xkcd
        >>> 'sopel.builtins.xkcd' in sys.modules
        True

    """

    PLUGIN_TYPE = 'python-module'
    """The plugin's type.

    Metadata for the plugin; this should be considered to be a constant and
    should not be modified at runtime.
    """

    def __init__(self, name, package=None):
        self.name = name
        self.package = package
        if package:
            self.module_name = self.package + '.' + self.name
        else:
            self.module_name = name

        self._module = None

    @property
    def module(self) -> ModuleType:
        if self._module is None:
            raise RuntimeError('No module for plugin %s' % self.name)
        return self._module

    def get_label(self):
        """Retrieve a display label for the plugin.

        :return: a human readable label for display purpose
        :rtype: str

        By default, this is ``<name> plugin``. If the plugin's module has a
        docstring, its first line is used as the plugin's label.
        """
        default_label = '%s plugin' % self.name

        if not self.is_loaded() or not hasattr(self.module, '__doc__'):
            return default_label

        module_doc = self.module.__doc__ or ""
        lines = inspect.cleandoc(module_doc).splitlines()
        return default_label if not lines else lines[0]

    def get_meta_description(self) -> PluginMetaDescription:
        """Retrieve a meta description for the plugin.

        :return: Metadata about the plugin
        :rtype: :class:`dict`

        The expected keys are detailed in :class:`PluginMetaDescription`.

        This implementation uses its module's dotted import path as the
        ``source`` value::

            {
                'name': 'example',
                'type': 'python-module',
                'label': 'example plugin',
                'source': 'sopel_modules.example',
                'version': '3.1.2',
            }

        """
        return {
            'label': self.get_label(),
            'type': self.PLUGIN_TYPE,
            'name': self.name,
            'source': self.module_name,
            'version': self.get_version(),
        }

    def get_version(self) -> Optional[str]:
        """Retrieve the plugin's version.

        :return: the plugin's version string
        :rtype: Optional[str]
        """
        version: Optional[str] = None
        if self.is_loaded() and hasattr(self.module, "__version__"):
            version = str(self.module.__version__)
        elif self.module_name.startswith("sopel."):
            version = release

        return version

    def load(self):
        """Load the plugin's module using :func:`importlib.import_module`.

        This method assumes the module is available through ``sys.path``.
        """
        self._module = importlib.import_module(self.module_name)

    def reload(self):
        """Reload the plugin's module using :func:`importlib.reload`.

        This method assumes the plugin is already loaded.
        """
        self._module = importlib.reload(self.module)

    def is_loaded(self):
        return self._module is not None

    def setup(self, bot):
        if self.has_setup():
            self.module.setup(bot)

    def has_setup(self):
        """Tell if the plugin has a setup action.

        :return: ``True`` if the plugin has a setup, ``False`` otherwise
        :rtype: bool

        The plugin has a setup action if its module has a ``setup`` attribute.
        This attribute is expected to be a callable.
        """
        return hasattr(self.module, 'setup')

    def get_capability_requests(self) -> list[plugin_decorators.capability]:
        return [
            module_attribute
            for module_attribute in vars(self.module).values()
            if isinstance(module_attribute, plugin_decorators.capability)
        ]

    def register(self, bot: Sopel) -> None:
        # capabilities are directly registered
        for cap_request in self.get_capability_requests():
            bot.cap_requests.register(self.name, cap_request)

        # plugin callables go through ``bot.add_plugin``
        relevant_parts = loader.clean_module(self.module, bot.config)
        for part in itertools.chain(*relevant_parts):
            # annotate all callables in relevant_parts with `plugin_name`
            # attribute to make per-channel config work; see #1839
            setattr(part, 'plugin_name', self.name)

        # TODO: replace add_plugin to direct call to register_* methods
        bot.add_plugin(self, *relevant_parts)

    def unregister(self, bot):
        relevant_parts = loader.clean_module(self.module, bot.config)
        bot.remove_plugin(self, *relevant_parts)

    def shutdown(self, bot):
        if self.has_shutdown():
            self.module.shutdown(bot)

    def has_shutdown(self):
        """Tell if the plugin has a shutdown action.

        :return: ``True`` if the plugin has a ``shutdown`` action, ``False``
                 otherwise
        :rtype: bool

        The plugin has a shutdown action if its module has a ``shutdown``
        attribute. This attribute is expected to be a callable.
        """
        return hasattr(self.module, 'shutdown')

    def configure(self, settings):
        if self.has_configure():
            self.module.configure(settings)

    def has_configure(self):
        """Tell if the plugin has a configure action.

        :return: ``True`` if the plugin has a ``configure`` action, ``False``
                 otherwise
        :rtype: bool

        The plugin has a configure action if its module has a ``configure``
        attribute. This attribute is expected to be a callable.
        """
        return hasattr(self.module, 'configure')


class PyFilePlugin(PyModulePlugin):
    """Sopel plugin loaded from the filesystem outside of the Python path.

    This plugin handler can be used to load a Sopel plugin from the
    filesystem, either a Python ``.py`` file or a directory containing an
    ``__init__.py`` file, and behaves like a :class:`PyModulePlugin`::

        >>> from sopel.plugins.handlers import PyFilePlugin
        >>> plugin = PyFilePlugin('/home/sopel/.sopel/plugins/custom.py')
        >>> plugin.load()
        >>> plugin.name
        'custom'

    In this example, the plugin ``custom`` is loaded from its filename despite
    not being in the Python path.
    """

    PLUGIN_TYPE = 'python-file'
    """The plugin's type.

    Metadata for the plugin; this should be considered to be a constant and
    should not be modified at runtime.
    """

    def __init__(self, filename):
        good_file = (
            os.path.isfile(filename) and
            filename.endswith('.py') and not filename.startswith('_')
        )
        good_dir = (
            os.path.isdir(filename) and
            os.path.isfile(os.path.join(filename, '__init__.py'))
        )

        if good_file:
            name = os.path.basename(filename)[:-3]
            spec = importlib.util.spec_from_file_location(
                name,
                filename,
            )
        elif good_dir:
            name = os.path.basename(filename)
            spec = importlib.util.spec_from_file_location(
                name,
                os.path.join(filename, '__init__.py'),
                submodule_search_locations=[filename],
            )
        else:
            raise exceptions.PluginError('Invalid Sopel plugin: %s' % filename)

        if spec is None:
            raise exceptions.PluginError('Could not determine spec for plugin: %s' % filename)

        self.filename = filename
        self.path = filename
        self.module_spec = spec

        super().__init__(name)

    def _load(self):
        module = importlib.util.module_from_spec(self.module_spec)
        if not self.module_spec.loader:
            raise exceptions.PluginError('Could not determine loader for plugin: %s' % self.filename)
        sys.modules[self.name] = module
        self.module_spec.loader.exec_module(module)
        return module

    def get_meta_description(self) -> PluginMetaDescription:
        """Retrieve a meta description for the plugin.

        :return: Metadata about the plugin
        :rtype: :class:`dict`

        The expected keys are detailed in :class:`PluginMetaDescription`.

        This implementation uses its source file's path as the ``source``
        value::

            {
                'name': 'example',
                'type': 'python-file',
                'label': 'example plugin',
                'source': '/home/username/.sopel/plugins/example.py',
                'version': '3.1.2',
            }

        """
        data = super().get_meta_description()
        data.update({
            'source': self.path,
        })
        return data

    def load(self):
        self._module = self._load()

    def reload(self):
        """Reload the plugin.

        Unlike :class:`PyModulePlugin`, it is not possible to use the
        ``reload`` function (either from `imp` or `importlib`), because the
        module might not be available through ``sys.path``.
        """
        self._module = self._load()


class EntryPointPlugin(PyModulePlugin):
    """Sopel plugin loaded from an entry point.

    :param entry_point: an entry point object

    This handler loads a Sopel plugin exposed by a package's entry point. It
    expects to be able to load a module object from the entry point, and to
    work as a :class:`~.PyModulePlugin` from that module.

    By default, Sopel searches within the entry point group ``sopel.plugins``.
    To use that for their own plugins, developers must define an entry point
    either in their ``setup.py`` file or their ``setup.cfg`` file::

        # in setup.py file
        setup(
            name='my_plugin',
            version='1.0',
            entry_points={
                'sopel.plugins': [
                    'custom = my_plugin.path.to.plugin',
                ],
            }
        )

    And this plugin can be loaded with::

        >>> from importlib_metadata import entry_points
        >>> from sopel.plugins.handlers import EntryPointPlugin
        >>> plugin = [
        ...     EntryPointPlugin(ep)
        ...     for ep in entry_points(group='sopel.plugins', name='custom')
        ... ][0]
        >>> plugin.load()
        >>> plugin.name
        'custom'

    In this example, the plugin ``custom`` is loaded from an entry point.
    Unlike the :class:`~.PyModulePlugin`, the name is not derived from the
    actual Python module, but from its entry point's name.

    .. seealso::

        Sopel uses the :func:`~sopel.plugins.find_entry_point_plugins` function
        internally to search entry points.

        Entry points are a `standard packaging mechanism`__ for Python, used by
        other applications (such as ``pytest``) for their plugins.

        The ``importlib_metadata`` backport package is used for consistency
        across all of Sopel's supported Python versions. Its API matches that
        of :mod:`importlib.metadata` from Python 3.10 and up; Sopel will drop
        this external requirement when practical.

        .. __: https://packaging.python.org/en/latest/specifications/entry-points/

    """

    PLUGIN_TYPE = 'setup-entrypoint'
    """The plugin's type.

    Metadata for the plugin; this should be considered to be a constant and
    should not be modified at runtime.
    """

    def __init__(self, entry_point):
        self.entry_point = entry_point
        super().__init__(entry_point.name)

    def load(self):
        self._module = self.entry_point.load()

    def get_version(self) -> Optional[str]:
        """Retrieve the plugin's version.

        :return: the plugin's version string
        :rtype: Optional[str]
        """
        version: Optional[str] = super().get_version()

        if (
            version is None
            and hasattr(self.entry_point, "dist")
            and hasattr(self.entry_point.dist, "name")
        ):
            dist_name = self.entry_point.dist.name
            try:
                version = importlib.metadata.version(dist_name)
            except (ValueError, importlib.metadata.PackageNotFoundError):
                LOGGER.warning("Cannot determine version of %r", dist_name)
            except Exception:
                LOGGER.warning(
                    "Unexpected error occurred while checking the version of %r",
                    dist_name,
                    exc_info=True,
                )

        return version

    def get_meta_description(self) -> PluginMetaDescription:
        """Retrieve a meta description for the plugin.

        :return: Metadata about the plugin
        :rtype: :class:`dict`

        The expected keys are detailed in :class:`PluginMetaDescription`.

        This implementation uses its entry point definition as the ``source``
        value::

            {
                'name': 'example',
                'type': 'setup-entrypoint',
                'label': 'example plugin',
                'source': 'example = my_plugin.example',
                'version': '3.1.2',
            }

        """
        data = super().get_meta_description()
        data.update({
            'source': self.entry_point.name + ' = ' + self.entry_point.value,
        })
        return data
